# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from . import SimpleTarget

LOG = logging.getLogger(__name__)


class FDBTarget(SimpleTarget):
    def __init__(self, fdb=None, config=None, userconfig=None, **kwargs):
        super().__init__(**kwargs)
        self._fdb = fdb
        self._fdb_kwargs = {}
        if config is not None:
            self._fdb_kwargs["config"] = config
        if userconfig is not None:
            self._fdb_kwargs["userconfig"] = userconfig

    @property
    def fdb(self):
        if self._fdb is None:
            import pyfdb

            self._fdb = pyfdb.FDB(**self._fdb_kwargs)
        return self._fdb

    def close(self):
        """Close the target and flush the fdb.

        The target will not be able to write anymore.

        Raises:
        -------
        ValueError: If the target is already closed.
        """
        self.flush()

    def flush(self):
        """Flush the fdb.

        Raises:
        -------
        ValueError: If the target is already closed.
        """
        self._raise_if_closed()
        self.fdb.flush()

    def _write(self, data, **kwargs):
        r = self._encode(data, **kwargs)
        if hasattr(r, "__iter__"):
            for d in r:
                self.fdb.archive(d.to_bytes())
        else:
            self.fdb.archive(r.to_bytes())


target = FDBTarget



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data.utils import ensure_iterable
from earthkit.data.utils import ensure_sequence


@pytest.mark.parametrize(
    "data,expected",
    [("a", ["a"]), ([1], [1]), ((1,), (1,)), (1, [1]), ({"a": 1}, {"a": 1})],
)
def test_utils_ensure_iterable(data, expected):
    assert ensure_iterable(data) == expected


@pytest.mark.parametrize(
    "data,expected",
    [
        ("a", ["a"]),
        ("abc", ["abc"]),
        ([1], [1]),
        ((1,), (1,)),
        (1, [1]),
        ({"a": 1}, [{"a": 1}]),
    ],
)
def test_utils_ensure_sequence(data, expected):
    assert ensure_sequence(data) == expected, f"{data=} {expected=}"


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import pytest

from earthkit.data import from_source


@pytest.fixture
def lod():
    prototype = {
        "gridType": "regular_ll",
        "Nx": 2,
        "Ny": 3,
        "distinctLatitudes": [-10.0, 0.0, 10.0],
        "distinctLongitudes": [0.0, 10.0],
        "_param_id": 167,
        "values": [[1, 2], [3, 4], [5, 6]],
        "date": "20180801",
        "time": "1200",
    }
    return [
        {"param": "t", "levelist": 500, **prototype},
        {"param": "t", "levelist": 850, **prototype},
        {"param": "u", "levelist": 500, **prototype},
        {"param": "u", "levelist": 850, **prototype},
        {"param": "d", "levelist": 850, **prototype},
        {"param": "d", "levelist": 600, **prototype},
    ]


def test_list_of_dicts(lod):
    ds = from_source("list-of-dicts", lod)

    assert len(ds) == 6
    ref = [("t", 500), ("t", 850), ("u", 500), ("u", 850), ("d", 850), ("d", 600)]
    assert ds.metadata("param", "levelist") == ref

    assert ds[0].metadata("step", default=None) is None

    assert ds[0].datetime() == {
        "base_time": datetime.datetime(2018, 8, 1, 12, 0),
        "valid_time": datetime.datetime(2018, 8, 1, 12, 0),
    }


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import shutil

try:
    import pyfdb
except ImportError:
    raise ImportError("FDB access requires 'pyfdb' to be installed")

from earthkit.data.sources.file import FileSource
from earthkit.data.sources.stream import StreamSource
from earthkit.data.utils.request import RequestMapper

from . import Source

LOG = logging.getLogger(__name__)


class FDBSource(Source):
    def __init__(self, *args, stream=True, config=None, userconfig=None, lazy=False, **kwargs):
        super().__init__()

        for k in ["group_by", "batch_size"]:
            if k in kwargs:
                raise ValueError(f"Invalid argument '{k}' for FDBSource. Deprecated since 0.8.0.")

        self.lazy = lazy
        self._fdb_kwargs = {}
        if config is not None:
            self._fdb_kwargs["config"] = config
        if userconfig is not None:
            self._fdb_kwargs["userconfig"] = userconfig

        self._stream_kwargs = dict()
        for k in ["read_all"]:
            if k in kwargs:
                self._stream_kwargs[k] = kwargs.pop(k)

        self.stream = stream

        self.request = {}
        for a in args:
            self.request.update(a)
        self.request.update(kwargs)

        if not (config or userconfig):
            self._check_env()

    def _check_env(self):
        fdb_home = os.environ.get("FDB_HOME", None)
        fdb_conf = os.environ.get("FDB5_CONFIG", None)
        if fdb_home is None and fdb_conf is None:
            raise ValueError(
                """Neither FDB_HOME nor FDB5_CONFIG environment variable
                was set! Please define either one to access FDB.
                See: https://fields-database.readthedocs.io for details about FDB."""
            )

    def mutate(self):
        if not self.lazy:
            fdb = pyfdb.FDB(**self._fdb_kwargs)
            if self.stream:
                stream = fdb.retrieve(self.request)
                return StreamSource(stream, **self._stream_kwargs)
            else:
                return FDBFileSource(fdb, self.request)
        else:
            mapper = FDBRequestMapper(self.request, fdb_kwargs=self._fdb_kwargs)
            retriever = FDBRetriever(self._fdb_kwargs)
            from earthkit.data.readers.grib.virtual import VirtualGribFieldList

            return VirtualGribFieldList(mapper, retriever)


class FDBFileSource(FileSource):
    def __init__(self, fdb, request):
        super().__init__()
        self.fdb = fdb
        self.path = self._retrieve(request)

    def _retrieve(self, request):
        def retrieve(target, request):
            with open(target, "wb") as o, self.fdb.retrieve(request) as i:
                shutil.copyfileobj(i, o)

        return self.cache_file(
            retrieve,
            request,
        )


class FDBRetriever:
    def __init__(self, fdb_kwargs):
        self.fdb_kwargs = fdb_kwargs

    def get(self, request):
        from . import from_source

        return from_source("fdb", request, stream=True, read_all=True, **self.fdb_kwargs)


class FDBRequestMapper(RequestMapper):
    def __init__(self, request, fdb_kwargs=None, **kwargs):
        super().__init__(request, **kwargs)
        self.fdb_kwargs = fdb_kwargs or {}
        self.metadata_alias = {
            "stepRange": "step",
            "typeOfLevel": "leveltype",
            "level": "levelist",
            "dataDate": "date",
            "dataTime": "time",
        }

    def _build(self):
        r = []
        fdb = pyfdb.FDB(**self.fdb_kwargs)
        for el in fdb.list(self.request, True, True):
            r.append(el["keys"])
        return r


source = FDBSource



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import sys

import pytest

from earthkit.data import from_source
from earthkit.data.testing import NO_POLYTOPE


def test_no_polytope_client(monkeypatch):
    "Check that a useful message is given in the absence of the polytope-client library"
    monkeypatch.setitem(sys.modules, "polytope", None)
    with pytest.raises(ImportError) as excinfo:
        from_source("polytope", None, None)
        assert "pip install polytope-client" in str(excinfo.value)


@pytest.mark.skip("Data not available?")
@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_POLYTOPE, reason="No access to Polytope Web API")
def test_polytope_odb():
    request = {
        "database": "fdbdev",
        "class": "rd",
        "type": "oai",
        "stream": "lwda",
        "expver": "xxxx",
        "obsgroup": "CONV",
        "reportype": 16001,
        "obstype": 15,
        "date": [20150601, 20150602],
        "time": "0/to/23/by/1",
        "filter": "'select * where stationid=\"Ams01\"'",
        "domain": "off",
    }

    src = from_source("polytope", "ichange", request, stream=False)
    df = src.to_pandas()
    assert len(df) == 52


# @pytest.mark.skip("Data not available?")
@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_POLYTOPE, reason="No access to Polytope Web API")
@pytest.mark.parametrize("kwargs", [{"stream": False}, {"stream": True, "read_all": True}])
def test_polytope_grib_all(kwargs):
    request = {
        "stream": "oper",
        "levtype": "pl",
        "levellist": "500",
        "param": "129.128",
        "step": "0/12",
        "time": "00:00:00",
        "date": "20200915",
        "type": "fc",
        "class": "rd",
        "expver": "hsvs",
        "domain": "g",
    }

    ds = from_source("polytope", "ecmwf-mars", request, **kwargs)

    assert len(ds) == 2
    assert ds.metadata("level") == [500, 500]


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_POLYTOPE, reason="No access to Polytope Web API")
def test_polytope_grib_stream():
    request = {
        "stream": "oper",
        "levtype": "pl",
        "levellist": "500",
        "param": "129.128",
        "step": "0/12",
        "time": "00:00:00",
        "date": "20200915",
        "type": "fc",
        "class": "rd",
        "expver": "hsvs",
        "domain": "g",
    }

    ds = from_source("polytope", "ecmwf-mars", request, stream=True)

    # no fieldlist methods are available
    with pytest.raises((TypeError, NotImplementedError)):
        len(ds)

    ref = [
        ("z", 500),
        ("z", 500),
    ]
    cnt = 0
    for i, f in enumerate(ds):
        assert f.metadata(("param", "level")) == ref[i], i
        cnt += 1

    assert cnt == len(ref)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


try:
    import cdsapi
except ImportError:
    raise ImportError("ADS access requires 'cdsapi' to be installed")

import yaml

from .cds import CdsRetriever
from .prompt import APIKeyPrompt


class ADSAPIKeyPrompt(APIKeyPrompt):
    register_or_sign_in_url = "https://ads.atmosphere.copernicus.eu/"
    retrieve_api_key_url = "https://ads.atmosphere.copernicus.eu/how-to-api"

    prompts = [
        dict(
            name="url",
            default="https://ads.atmosphere.copernicus.eu/api",
            title="API url",
            validate=r"http.?://.*",
        ),
        dict(
            name="key",
            example="abcdef01-0000-1111-2222-0123456789ab",
            title="API key",
            hidden=True,
            validate=r"[\-0-9a-f]+",
        ),
    ]

    rcfile = "~/.adsapirc"
    rc_message_base = """You have to store the credentials in {rcfile}; if you follow the
           instructions below it will be automatically done for you.
    """

    def save(self, input, file):
        yaml.dump(input, file, default_flow_style=False)

    def load(self, file):
        return yaml.safe_load(file.read())


def client():
    prompt = ADSAPIKeyPrompt()
    rc = prompt.check(load=True)
    return cdsapi.Client(**rc)


EXTENSIONS = {
    "grib": ".grib",
    "netcdf": ".nc",
}


class AdsRetriever(CdsRetriever):
    sphinxdoc = """
    AdsRetriever
    """

    def client(self):
        return client()


source = AdsRetriever



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_directory
from earthkit.data.core.temporary import temp_file
from earthkit.data.sources.stream import StreamFieldList
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_remote_test_data_file
from earthkit.data.testing import make_tgz
from earthkit.data.testing import write_to_file


def repeat_list_items(items, count):
    return sum([[x] * count for x in items], [])


def test_grib_file_stream_iter():
    ds = from_source(
        "file",
        earthkit_examples_file("test6.grib"),
        stream=True,
    )

    # no fieldlist methods are available
    with pytest.raises((TypeError, NotImplementedError)):
        len(ds)

    ref = [
        ("t", 1000),
        ("u", 1000),
        ("v", 1000),
        ("t", 850),
        ("u", 850),
        ("v", 850),
    ]
    cnt = 0
    for i, f in enumerate(ds):
        assert f.metadata(("param", "level")) == ref[i], i
        cnt += 1

    assert cnt == len(ref)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize(
    "_kwargs,expected_meta",
    [
        ({"n": 1}, [["t"], ["u"], ["v"], ["t"], ["u"], ["v"]]),
        ({"n": 3}, [["t", "u", "v"], ["t", "u", "v"]]),
        ({"n": 4}, [["t", "u", "v", "t"], ["u", "v"]]),
    ],
)
def test_grib_file_stream_batched(_kwargs, expected_meta):
    ds = from_source(
        "file",
        earthkit_examples_file("test6.grib"),
        stream=True,
    )

    # no methods are available
    with pytest.raises((TypeError, NotImplementedError)):
        len(ds)

    cnt = 0
    for i, f in enumerate(ds.batched(_kwargs["n"])):
        assert len(f) == len(expected_meta[i])
        f.metadata("param") == expected_meta[i]
        cnt += 1

    assert cnt == len(expected_meta)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize("group", ["level", ["level", "gridType"]])
def test_grib_file_stream_group_by(group):
    ds = from_source("file", earthkit_examples_file("test6.grib"), stream=True)

    # no methods are available
    with pytest.raises((TypeError, NotImplementedError)):
        len(ds)

    ref = [
        [("t", 1000), ("u", 1000), ("v", 1000)],
        [("t", 850), ("u", 850), ("v", 850)],
    ]
    cnt = 0
    for i, f in enumerate(ds.group_by(group)):
        assert len(f) == 3
        assert f.metadata(("param", "level")) == ref[i]
        assert f.to_fieldlist(array_backend="numpy") is not f
        cnt += 1

    assert cnt == len(ref)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


def test_grib_file_stream_in_memory():
    ds = from_source(
        "file",
        earthkit_examples_file("test6.grib"),
        stream=True,
        read_all=True,
    )

    assert len(ds) == 6

    expected_shape = (6, 7, 12)
    ref = ["t", "u", "v", "t", "u", "v"]

    # iteration
    val = [f.metadata(("param")) for f in ds]
    assert val == ref, "iteration"

    # metadata
    val = ds.metadata("param")
    assert val == ref, "method"

    # data
    assert ds.to_numpy().shape == expected_shape

    ref = np.array(
        [
            272.56417847,
            -6.28688049,
            7.83348083,
            272.53916931,
            -4.89837646,
            8.66096497,
        ]
    )

    vals = ds.to_numpy()[:, 0, 0]
    assert np.allclose(vals, ref)


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_save_when_loaded_from_file_stream(write_method):
    ds = from_source(
        "file",
        earthkit_examples_file("test6.grib"),
        stream=True,
        read_all=True,
    )
    assert len(ds) == 6
    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds)
        ds_saved = from_source("file", tmp)
        assert len(ds) == len(ds_saved)


# @pytest.mark.parametrize(
#     "_kwargs",
#     [
#         {},
#         {"batch_size": 1},
#     ],
# )
def test_grib_file_stream_multi_file_iter():
    ds = from_source(
        "file",
        [
            earthkit_examples_file("test.grib"),
            earthkit_examples_file("test4.grib"),
        ],
        stream=True,
    )

    assert isinstance(ds, StreamFieldList)
    assert len(ds._source.sources) == 2

    # no fieldlist methods are available
    with pytest.raises((TypeError, NotImplementedError)):
        len(ds)

    ref = [
        ("2t", 0),
        ("msl", 0),
        ("t", 500),
        ("z", 500),
        ("t", 850),
        ("z", 850),
    ]
    cnt = 0
    for i, f in enumerate(ds):
        assert f.metadata(("param", "level")) == ref[i], i
        cnt += 1

    assert cnt == len(ref)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize(
    "_kwargs,expected_meta",
    [
        ({"n": 1}, [["2t"], ["msl"], ["t"], ["z"], ["t"], ["z"]]),
        ({"n": 2}, [["2t", "msl"], ["t", "z"], ["t", "z"]]),
        ({"n": 3}, [["2t", "msl", "t"], ["z", "t", "z"]]),
        ({"n": 4}, [["2t", "msl", "t", "z"], ["t", "z"]]),
    ],
)
def test_grib_file_stream_multi_file_batched(_kwargs, expected_meta):
    ds = from_source(
        "file",
        [
            earthkit_examples_file("test.grib"),
            earthkit_examples_file("test4.grib"),
        ],
        stream=True,
    )

    # no methods are available
    with pytest.raises((TypeError, NotImplementedError)):
        len(ds)

    cnt = 0
    for i, f in enumerate(ds.batched(_kwargs["n"])):
        assert len(f) == len(expected_meta[i])
        f.metadata("param") == expected_meta[i]
        cnt += 1

    assert cnt == len(expected_meta)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


def test_grib_file_stream_multi_file_memory():
    ds = from_source(
        "file",
        [
            earthkit_examples_file("test.grib"),
            earthkit_examples_file("test4.grib"),
        ],
        stream=True,
        read_all=True,
    )

    assert len(ds) == 6

    md_ref = [
        ("2t", 0),
        ("msl", 0),
        ("t", 500),
        ("z", 500),
        ("t", 850),
        ("z", 850),
    ]
    # iteration
    val = [f.metadata(("param", "level")) for f in ds]
    assert val == md_ref, "iteration"

    # metadata
    val = ds.metadata(("param", "level"))
    assert val == md_ref, "method"

    # data
    with pytest.raises(ValueError):
        ds.to_numpy().shape

    # first part
    expected_shape = (2, 11, 19)
    assert ds[0:2].to_numpy().shape == expected_shape

    ref = np.array([262.78027344, 101947.8125])

    vals = ds[0:2].to_numpy()[:, 0, 0]
    assert np.allclose(vals, ref)

    # second part
    expected_shape = (4, 181, 360)
    assert ds[2:].to_numpy().shape == expected_shape

    ref = np.array([228.04600525, 48126.859375, 246.61032104, 11786.1132812])

    vals = ds[2:].to_numpy()[:, 0, 0]
    assert np.allclose(vals, ref)

    # slicing
    r = ds[0:3]
    assert len(r) == 3
    val = r.metadata(("param", "level"))
    assert val == md_ref[0:3]

    r = ds[-2:]
    assert len(r) == 2
    val = r.metadata(("param", "level"))
    assert val == md_ref[-2:]

    r = ds.sel(param="t")
    assert len(r) == 2
    val = r.metadata(("param", "level"))
    assert val == [
        ("t", 500),
        ("t", 850),
    ]


@pytest.mark.cache
@pytest.mark.parametrize(
    "path,parts,expected_meta,remote",
    [
        ("test6.grib", [(0, 150)], [("t", 1000)], False),
        ("test6.grib", [(240, 150)], [("u", 1000)], False),
        ("test6.grib", [(240, 480)], [("u", 1000), ("v", 1000)], False),
        ("test-data/karl_850.grib", [(0, 1683960)], [("t", 850)], True),
        ("test-data/karl_850.grib", [(0, 3367920)], [("t", 850), ("r", 850)], True),
        ("test6.grib", [(240, 240), (720, 240)], [("u", 1000), ("t", 850)], False),
        ("test-data/karl_850.grib", [(0, 1683960), (3367920, 1683960)], [("t", 850), ("z", 850)], True),
    ],
)
def test_grib_file_stream_single_file_parts_core(path, parts, expected_meta, remote):
    if remote:
        ds0 = from_source(
            "url",
            earthkit_remote_test_data_file(path),
        )
        path = ds0.path

    ds = from_source(
        "file",
        earthkit_examples_file(path),
        parts=parts,
        stream=True,
    )

    # no fieldlist methods are available
    with pytest.raises((TypeError, NotImplementedError)):
        len(ds)

    cnt = 0
    for i, f in enumerate(ds):
        assert f.metadata(("param", "level")) == expected_meta[i], i
        cnt += 1

    assert cnt == len(expected_meta)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize(
    "parts,expected_meta",
    [
        ([(0, 150)], [("t", 1000)]),
        (
            None,
            [("t", 1000), ("u", 1000), ("v", 1000), ("t", 850), ("u", 850), ("v", 850)],
        ),
    ],
)
def test_grib_file_stream_single_file_parts_as_arg_valid(parts, expected_meta):
    ds = from_source(
        "file",
        [earthkit_examples_file("test6.grib"), parts],
        stream=True,
    )

    # no fieldlist methods are available
    with pytest.raises((TypeError, NotImplementedError)):
        len(ds)

    cnt = 0
    for i, f in enumerate(ds):
        assert f.metadata(("param", "level")) == expected_meta[i], i
        cnt += 1

    assert cnt == len(expected_meta)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


def test_grib_file_stream_single_file_parts_as_arg_invalid():
    with pytest.raises(ValueError):
        from_source(
            "file",
            [earthkit_examples_file("test6.grib"), [(0, 150)]],
            parts=[(0, 160)],
            stream=True,
        )


@pytest.mark.parametrize(
    "parts1,parts2,expected_meta",
    [
        (
            [(240, 150)],
            None,
            [("u", 1000), ("2t", 0), ("msl", 0)],
        ),
        (
            None,
            [(0, 526)],
            [
                ("t", 1000),
                ("u", 1000),
                ("v", 1000),
                ("t", 850),
                ("u", 850),
                ("v", 850),
                ("2t", 0),
            ],
        ),
        (
            [(240, 150)],
            [(0, 526)],
            [("u", 1000), ("2t", 0)],
        ),
        (
            [(240, 150), (720, 150)],
            [(0, 526)],
            [("u", 1000), ("t", 850), ("2t", 0)],
        ),
    ],
)
def test_grib_file_stream_multi_file_parts(parts1, parts2, expected_meta):
    ds = from_source(
        "file",
        [
            [earthkit_examples_file("test6.grib"), parts1],
            [earthkit_examples_file("test.grib"), parts2],
        ],
        stream=True,
    )

    # no fieldlist methods are available
    with pytest.raises((TypeError, NotImplementedError)):
        len(ds)

    cnt = 0
    for i, f in enumerate(ds):
        assert f.metadata(("param", "level")) == expected_meta[i], i
        cnt += 1

    assert cnt == len(expected_meta)

    # stream consumed, no data is available
    assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_file_stream_glob(write_method):
    s1 = from_source("file", earthkit_examples_file("test.grib"))
    s2 = from_source("file", earthkit_examples_file("test4.grib"))
    with temp_directory() as tmpdir:
        write_to_file(write_method, os.path.join(tmpdir, "a.grib"), s1)
        write_to_file(write_method, os.path.join(tmpdir, "b.grib"), s2)

        ds = from_source("file", os.path.join(tmpdir, "*a.grib"), stream=True)

        # no fieldlist methods are available
        with pytest.raises((TypeError, NotImplementedError)):
            len(ds)

        ref = [
            ("2t", 0),
            ("msl", 0),
        ]
        cnt = 0
        for i, f in enumerate(ds):
            assert f.metadata(("param", "level")) == ref[i], i
            cnt += 1

        assert cnt == len(ref)

        # stream consumed, no data is available
        assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_file_stream_single_directory(write_method):
    s1 = from_source("file", earthkit_examples_file("test.grib"))
    s2 = from_source("file", earthkit_examples_file("test4.grib"))
    with temp_directory() as tmpdir:
        write_to_file(write_method, os.path.join(tmpdir, "a.grib"), s1)
        write_to_file(write_method, os.path.join(tmpdir, "b.grib"), s2)

        ds = from_source("file", tmpdir, stream=True)

        # no fieldlist methods are available
        with pytest.raises((TypeError, NotImplementedError)):
            len(ds)

        ref = [
            ("2t", 0),
            ("msl", 0),
            ("t", 500),
            ("z", 500),
            ("t", 850),
            ("z", 850),
        ]
        cnt = 0
        for i, f in enumerate(ds):
            assert f.metadata(("param", "level")) == ref[i], i
            cnt += 1

        assert cnt == len(ref)

        # stream consumed, no data is available
        assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_file_stream_multi_directory(write_method):
    s1 = from_source("file", earthkit_examples_file("test.grib"))
    s2 = from_source("file", earthkit_examples_file("test4.grib"))
    s3 = from_source("file", earthkit_examples_file("test6.grib"))
    with temp_directory() as tmpdir1:
        write_to_file(write_method, os.path.join(tmpdir1, "a.grib"), s1)
        write_to_file(write_method, os.path.join(tmpdir1, "b.grib"), s2)

        with temp_directory() as tmpdir2:
            write_to_file(write_method, os.path.join(tmpdir2, "a.grib"), s1)
            write_to_file(write_method, os.path.join(tmpdir2, "b.grib"), s3)

            ds = from_source("file", [tmpdir1, tmpdir2], stream=True)

            ref = [
                ("2t", 0),
                ("msl", 0),
                ("t", 500),
                ("z", 500),
                ("t", 850),
                ("z", 850),
                ("2t", 0),
                ("msl", 0),
                ("t", 1000),
                ("u", 1000),
                ("v", 1000),
                ("t", 850),
                ("u", 850),
                ("v", 850),
            ]

            cnt = 0
            for i, f in enumerate(ds):
                assert f.metadata(("param", "level")) == ref[i], i
                cnt += 1

            assert cnt == len(ref)

            # stream consumed, no data is available
            assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize("filter_kwarg", [(lambda x: "b.grib" in x), ("*b.grib")])
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_file_stream_single_directory_filter(filter_kwarg, write_method):
    s1 = from_source("file", earthkit_examples_file("test.grib"))
    s2 = from_source("file", earthkit_examples_file("test4.grib"))
    with temp_directory() as tmpdir:
        write_to_file(write_method, os.path.join(tmpdir, "a.grib"), s1)
        write_to_file(write_method, os.path.join(tmpdir, "b.grib"), s2)

        ds = from_source("file", tmpdir, filter=filter_kwarg, stream=True)

        ref = [
            ("t", 500),
            ("z", 500),
            ("t", 850),
            ("z", 850),
        ]

        cnt = 0
        for i, f in enumerate(ds):
            assert f.metadata(("param", "level")) == ref[i], i
            cnt += 1

        assert cnt == len(ref)

        # stream consumed, no data is available
        assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize("filter_kwarg", [(lambda x: "b.grib" in x), ("*b.grib")])
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_file_stream_multi_directory_filter(filter_kwarg, write_method):
    s1 = from_source("file", earthkit_examples_file("test.grib"))
    s2 = from_source("file", earthkit_examples_file("test4.grib"))
    s3 = from_source("file", earthkit_examples_file("test6.grib"))
    with temp_directory() as tmpdir1:
        write_to_file(write_method, os.path.join(tmpdir1, "a.grib"), s1)
        write_to_file(write_method, os.path.join(tmpdir1, "b.grib"), s2)

        with temp_directory() as tmpdir2:
            write_to_file(write_method, os.path.join(tmpdir2, "a.grib"), s1)
            write_to_file(write_method, os.path.join(tmpdir2, "b.grib"), s3)

            ds = from_source("file", [tmpdir1, tmpdir2], filter=filter_kwarg, stream=True)

            ref = [
                ("t", 500),
                ("z", 500),
                ("t", 850),
                ("z", 850),
                ("t", 1000),
                ("u", 1000),
                ("v", 1000),
                ("t", 850),
                ("u", 850),
                ("v", 850),
            ]

            cnt = 0
            for i, f in enumerate(ds):
                assert f.metadata(("param", "level")) == ref[i], i
                cnt += 1

            assert cnt == len(ref)

            # stream consumed, no data is available
            assert sum([1 for _ in ds]) == 0


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_file_stream_multi_directory_with_tar(write_method):
    s1 = from_source("file", earthkit_examples_file("test.grib"))
    s2 = from_source("file", earthkit_examples_file("test4.grib"))
    s3 = from_source("file", earthkit_examples_file("test6.grib"))
    with temp_directory() as tmpdir1:
        write_to_file(write_method, os.path.join(tmpdir1, "a.grib"), s1)
        write_to_file(write_method, os.path.join(tmpdir1, "b.grib"), s2)

        with temp_directory() as tmpdir2:
            write_to_file(write_method, os.path.join(tmpdir2, "a.grib"), s1)
            write_to_file(write_method, os.path.join(tmpdir2, "b.grib"), s3)

            paths = [os.path.join(tmpdir2, f) for f in ["a.grib", "b.grib"]]
            make_tgz(tmpdir2, "test.tar.gz", paths)

            ds = from_source("file", [tmpdir1, tmpdir2], stream=True)

            ref = [
                ("2t", 0),
                ("msl", 0),
                ("t", 500),
                ("z", 500),
                ("t", 850),
                ("z", 850),
                ("2t", 0),
                ("msl", 0),
                ("t", 1000),
                ("u", 1000),
                ("v", 1000),
                ("t", 850),
                ("u", 850),
                ("v", 850),
                ("2t", 0),
                ("msl", 0),
                ("t", 1000),
                ("u", 1000),
                ("v", 1000),
                ("t", 850),
                ("u", 850),
                ("v", 850),
            ]

            cnt = 0
            for i, f in enumerate(ds):
                assert f.metadata(("param", "level")) == ref[i], i
                cnt += 1

            assert cnt == len(ref)

            # stream consumed, no data is available
            assert sum([1 for _ in ds]) == 0


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np

import earthkit.data
from earthkit.data.testing import earthkit_remote_test_data_file


def test_netcdf_fieldlist_string_coord():
    import xarray as xr

    a = xr.DataArray(
        [
            [[11, 12, 13], [21, 22, 23], [31, 32, 33]],
            [[14, 15, 16], [24, 25, 26], [34, 35, 36]],
        ],
        coords={"level": ["500", "700"], "x": [1, 2, 3], "y": [4, 5, 6]},
        name="dummyvar",
    )

    # sanity check
    assert np.allclose(a.sel(level="500", x=2).values, np.array([21, 22, 23]))

    ds = earthkit.data.from_object(a)

    assert ds
    assert len(ds) == 2
    assert ds.metadata("level") == ["500", "700"]

    x_ref = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    y_ref = np.array([[4, 5, 6], [4, 5, 6], [4, 5, 6]])
    assert np.allclose(ds[0].to_points()["x"], x_ref)
    assert np.allclose(ds[0].to_points()["y"], y_ref)


def test_netcdf_fieldlist_bounds():
    """Check if having non string values in bounds does not cause a crash"""
    import xarray as xr

    a = xr.DataArray(
        [
            [[11, 12, 13], [21, 22, 23], [31, 32, 33]],
            [[14, 15, 16], [24, 25, 26], [34, 35, 36]],
        ],
        coords={"level": [0, 0], "x": [1, 2, 3], "y": [4, 5, 6]},
        attrs={"bounds": [1, 2, 3]},
        name="dummyvar",
    )

    ds = earthkit.data.from_object(a)
    assert ds
    assert len(ds) == 2


def test_netcdf_fieldlist_ctime():
    ds = earthkit.data.from_source(
        "url",
        earthkit_remote_test_data_file("test-data", "zgrid_rhgmet_metop_200701_R_2305_0010.nc"),
    )

    assert len(ds) == 1506
    assert ds[0].metadata("valid_datetime") == "2007-01-16T00:00:00"
    assert ds[5].metadata("valid_datetime") == "2007-01-16T00:00:00"


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import itertools
import logging
from functools import cached_property

import numpy as np

from earthkit.data.core.fieldlist import Field
from earthkit.data.core.fieldlist import FieldList
from earthkit.data.core.index import MaskIndex
from earthkit.data.core.metadata import RawMetadata
from earthkit.data.decorators import cached_method
from earthkit.data.decorators import normalize
from earthkit.data.indexing.fieldlist import ClonedFieldCore
from earthkit.data.utils.dates import to_datetime

LOG = logging.getLogger(__name__)


class ForcingMetadata(RawMetadata):
    LS_KEYS = ["valid_datetime", "param"]

    def __init__(self, d, geography):
        self._geo = geography
        super().__init__(d)

    @property
    def geography(self):
        return self._geo

    def datetime(self):
        return {
            "base_time": self.base_datetime(),
            "valid_time": self.valid_datetime(),
        }

    def base_datetime(self):
        return None

    def valid_datetime(self):
        return datetime.datetime.fromisoformat(self["valid_datetime"])

    def step_timedelta(self):
        return datetime.timedelta()

    def ls_keys(self):
        return self.LS_KEYS


class ForcingMaker:
    def __init__(self, field):
        self.field = field
        self.shape = self.field.shape

    @cached_method
    def grid_points(self):
        d = self.field.to_latlon(flatten=True)
        return d["lat"], d["lon"]

    @cached_method
    def ecef_xyz(self):
        # https://en.wikipedia.org/wiki/Geographic_coordinate_conversion#From_geodetic_to_ECEF_coordinates
        # We assume that the Earth is a sphere of radius 1 so N(phi) = 1
        # We assume h = 0

        lat, lon = self.grid_points()

        phi = np.deg2rad(lat)
        lda = np.deg2rad(lon)

        cos_phi = np.cos(phi)
        cos_lda = np.cos(lda)
        sin_phi = np.sin(phi)
        sin_lda = np.sin(lda)

        x = cos_phi * cos_lda
        y = cos_phi * sin_lda
        z = sin_phi

        return x, y, z

    @cached_method
    def latitude_(self):
        return self.grid_points()[0]

    def latitude(self, date):
        return self.latitude_()

    @cached_method
    def cos_latitude_(self):
        return np.cos(np.deg2rad(self.grid_points()[0]))

    def cos_latitude(self, date):
        return self.cos_latitude_()

    @cached_method
    def sin_latitude_(self):
        return np.sin(np.deg2rad(self.grid_points()[0]))

    def sin_latitude(self, date):
        return self.sin_latitude_()

    @cached_method
    def longitude_(self):
        return self.grid_points()[1]

    def longitude(self, date):
        return self.longitude_()

    @cached_method
    def cos_longitude_(self):
        return np.cos(np.deg2rad(self.grid_points()[1]))

    def cos_longitude(self, date):
        return self.cos_longitude_()

    @cached_method
    def sin_longitude_(self):
        return np.sin(np.deg2rad(self.grid_points()[1]))

    def sin_longitude(self, date):
        return self.sin_longitude_()

    def ecef_x(self, date):
        return self.ecef_xyz()[0]

    def ecef_y(self, date):
        return self.ecef_xyz()[1]

    def ecef_z(self, date):
        return self.ecef_xyz()[2]

    def julian_day(self, date):
        date = to_datetime(date)
        if date.tzinfo is not None and date.tzinfo.utcoffset(date) is not None:
            year_start = datetime.datetime(date.year, 1, 1, tzinfo=date.tzinfo)
        else:
            year_start = datetime.datetime(date.year, 1, 1)
        delta = date - year_start
        julian_day = delta.days + delta.seconds / 86400.0
        return np.full((np.prod(self.field.shape),), julian_day)

    def cos_julian_day(self, date):
        radians = self.julian_day(date) / 365.25 * np.pi * 2
        return np.cos(radians)

    def sin_julian_day(self, date):
        radians = self.julian_day(date) / 365.25 * np.pi * 2
        return np.sin(radians)

    def local_time(self, date):
        lon = self.longitude(date)
        date = to_datetime(date)
        if date.tzinfo is not None and date.tzinfo.utcoffset(date) is not None:
            day_start = datetime.datetime(date.year, date.month, date.day, tzinfo=date.tzinfo)
        else:
            day_start = datetime.datetime(date.year, date.month, date.day)
        delta = date - day_start
        hours_since_midnight = (delta.days + delta.seconds / 86400.0) * 24
        return (lon / 360.0 * 24.0 + hours_since_midnight) % 24

    def cos_local_time(self, date):
        radians = self.local_time(date) / 24 * np.pi * 2
        return np.cos(radians)

    def sin_local_time(self, date):
        radians = self.local_time(date) / 24 * np.pi * 2
        return np.sin(radians)

    def insolation(self, date):
        return self.cos_solar_zenith_angle(date)

    def toa_incident_solar_radiation(self, date):
        from earthkit.meteo.solar import toa_incident_solar_radiation

        date = to_datetime(date)
        result = toa_incident_solar_radiation(
            date - datetime.timedelta(minutes=30),
            date + datetime.timedelta(minutes=30),
            self.latitude_(),
            self.longitude_(),
            intervals_per_hour=2,
        )
        return result.flatten()

    def cos_solar_zenith_angle(self, date):
        from earthkit.meteo.solar import cos_solar_zenith_angle

        date = to_datetime(date)
        result = cos_solar_zenith_angle(
            date,
            self.latitude_(),
            self.longitude_(),
        )
        return result.flatten()

    def __getattr__(self, name):
        if "+" not in name and "-" not in name:
            # If we are here, we are looking for a method that does not exist,
            # it has to be a method with a time delta.
            raise AttributeError(name)
        if "+" in name:
            fname, delta = name.split("+")
            sign = 1
        if "-" in name:
            fname, delta = name.split("-")
            sign = -1
        method = getattr(self, fname)

        if delta.endswith("h"):
            factor = 60
        elif delta.endswith("d"):
            factor = 24 * 60
        else:
            raise ValueError(f"Invalid time delta {delta} in {name}")

        delta = delta[:-1]
        delta = int(delta)
        delta = datetime.timedelta(minutes=delta) * factor * sign

        def wrapper(date):
            date = date + delta
            value = method(date)
            return value

        return wrapper


class ForcingField(Field):
    def __init__(self, maker, date, param, proc, number=None):
        self.maker = maker
        self.date = date
        self.param = param
        self.proc = proc
        self.number = number
        # self._shape = shape
        # self._geometry = self.maker.field.metadata().geography

    @cached_property
    def _metadata(self):
        d = dict(
            valid_datetime=self.date if isinstance(self.date, str) else self.date.isoformat(),
            param=self.param,
            level=None,
            levelist=None,
            number=self.number,
            levtype=None,
        )
        return ForcingMetadata(d, self.maker.field.metadata().geography)

    def _values(self, dtype=None):
        values = self.proc(self.date)
        if dtype is not None:
            values = values.astype(dtype)
        return values

    def clone(self, **kwargs):
        return ClonedForcingField(self, **kwargs)

    def __repr__(self):
        return "ForcingField(%s,%s,%s)" % (self.param, self.date, self.number)


class ClonedForcingField(ClonedFieldCore, ForcingField):
    def __init__(self, field, **kwargs):
        ClonedFieldCore.__init__(self, field, **kwargs)
        ForcingField.__init__(self, field.maker, field.date, field.param, field.proc, number=field.number)


def make_datetime(date, time):
    if time is None:
        return date
    if date.hour or date.minute:
        raise ValueError(
            (
                f"Duplicate information about time time={time},"
                f"and time={date.hour}:{date.minute} from date={date}"
            )
        )
    assert date.hour == 0, (date, time)
    assert date.minute == 0, (date, time)
    assert str(time).isdigit(), (type(time), time)
    time = int(time)
    if time > 24:
        time = time // 100
    return datetime.datetime(date.year, date.month, date.day, time)


def index_to_coords(index: int, shape):
    assert isinstance(index, int), (index, type(index))

    result = [None] * len(shape)
    i = len(shape) - 1

    while i >= 0:
        result[i] = index % shape[i]
        index = index // shape[i]
        i -= 1

    result = tuple(result)

    assert len(result) == len(shape)
    return result


class ForcingsFieldListCore(FieldList):
    def __init__(self, source_or_dataset=None, *, request={}, **kwargs):
        request = dict(**request)
        request.update(kwargs)

        self.request = self._request(**request)

        def find_latlon():
            lats = None
            for k in ["latitudes", "latitude"]:
                if k in self.request:
                    lats = np.asarray(self.request.pop(k))
                    break

            lons = None
            for k in ["longitudes", "longitude"]:
                if k in self.request:
                    lons = np.asarray(self.request.pop(k))
                    break

            return lats, lons

        def find_numbers(source_or_dataset):
            if "number" in self.request:
                return self.request["number"]

            assert hasattr(source_or_dataset, "unique_values"), (
                f"{source_or_dataset} (type '{type(source_or_dataset).__name__}') is"
                " not a proper source or dataset"
            )

            return source_or_dataset.unique_values("number", patches={"number": {None: 0}}).get("number", 0)

        def find_dates(source_or_dataset):
            if "date" not in self.request and "time" in self.request:
                raise ValueError("Cannot specify time without date")

            if "date" in self.request and "time" not in self.request:
                return self.request["date"]

            if "date" in self.request and "time" in self.request:
                dates = [
                    make_datetime(date, time)
                    for date, time in itertools.product(self.request["date"], self.request["time"])
                ]
                assert len(set(dates)) == len(dates), "Duplicates dates in forcings."
                return dates

            assert "date" not in self.request and "time" not in self.request
            assert hasattr(source_or_dataset, "unique_values"), (
                f"{source_or_dataset} (type '{type(source_or_dataset).__name__}') is"
                " not a proper source or dataset"
            )

            return source_or_dataset.unique_values("valid_datetime")["valid_datetime"]

        if source_or_dataset is None:
            lats, lons = find_latlon()
            if lats is None:
                raise ValueError("latitudes must be specified when no source or dataset provided")

            if lons is None:
                raise ValueError("longitudes must be specified when no source or dataset provided")

            from earthkit.data import from_source

            vals = np.ones(lats.shape)
            d = {"values": vals, "latitudes": lats, "longitudes": lons}
            # d.update(self.request)

            source_or_dataset = from_source("list-of-dicts", [d])

        self.dates = find_dates(source_or_dataset)

        self.params = self.request["param"]
        if not isinstance(self.params, (tuple, list)):
            self.params = [self.params]

        # self.numbers = self.request.get("number", [None])
        self.numbers = find_numbers(source_or_dataset)
        if not isinstance(self.numbers, (tuple, list)):
            self.numbers = [self.numbers]

        self.maker = ForcingMaker(field=source_or_dataset[0])
        self.procs = {param: getattr(self.maker, param) for param in self.params}
        self._len = len(self.dates) * len(self.params) * len(self.numbers)

        super().__init__(**kwargs)

    @normalize("date", "date-list")
    @normalize("time", "int-list")
    @normalize("number", "int-list")
    def _request(self, **request):
        return request

    @classmethod
    def new_mask_index(self, *args, **kwargs):
        return ForcingsMaskFieldList(*args, **kwargs)


class ForcingsFieldList(ForcingsFieldListCore):
    def __len__(self):
        return self._len

    def _getitem(self, n):
        if isinstance(n, int):
            if n < 0:
                n = self._len + n

            if n >= self._len or n < 0:
                raise IndexError(n)

            date, param, number = index_to_coords(n, (len(self.dates), len(self.params), len(self.numbers)))

            date = self.dates[date]
            # assert isinstance(date, datetime.datetime), (date, type(date))
            param = self.params[param]
            number = self.numbers[number]

            return ForcingField(
                self.maker,
                date,
                param,
                self.procs[param],
                number=number,
            )


class ForcingsMaskFieldList(ForcingsFieldListCore, MaskIndex):
    def __init__(self, *args, **kwargs):
        MaskIndex.__init__(self, *args, **kwargs)
        FieldList._init_from_mask(self, self)


source = ForcingsFieldList



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
from functools import cached_property

import eccodes
import numpy as np

from earthkit.data.core.fieldlist import Field
from earthkit.data.indexing.fieldlist import ClonedFieldCore
from earthkit.data.readers.grib.metadata import GribFieldMetadata
from earthkit.data.utils.message import CodesHandle
from earthkit.data.utils.message import CodesMessagePositionIndex
from earthkit.data.utils.message import CodesReader

LOG = logging.getLogger(__name__)

_GRIB_NAMESPACES = {"default": None}

for k in ("ls", "geography", "mars", "parameter", "statistics", "time", "vertical"):
    _GRIB_NAMESPACES[k] = k


def missing_is_none(x):
    return None if x == 2147483647 else x


class GribCodesFloatArrayAccessor:
    HAS_FLOAT_SUPPORT = None
    KEY = None

    def __init__(self):
        if GribCodesFloatArrayAccessor.HAS_FLOAT_SUPPORT is None:
            GribCodesFloatArrayAccessor.HAS_FLOAT_SUPPORT = hasattr(eccodes, "codes_get_float_array")

    def get(self, handle, dtype=None):
        v = eccodes.codes_get_array(handle, self.KEY)
        if dtype is not None:
            return v.astype(dtype)
        else:
            return v

    @staticmethod
    def to_numpy_dtype(dtype):
        from earthkit.utils.array.dtype import to_numpy_dtype

        return to_numpy_dtype(dtype, default=np.float64)


class GribCodesValueAccessor(GribCodesFloatArrayAccessor):
    KEY = "values"

    def __init__(self):
        super().__init__()

    def get(self, handle, dtype=None):
        dtype = self.to_numpy_dtype(dtype)
        if dtype is np.float32 and self.HAS_FLOAT_SUPPORT:
            return eccodes.codes_get_array(handle, self.KEY, ktype=dtype)
        else:
            return super().get(handle, dtype=dtype)


class GribCodesLatitudeAccessor(GribCodesFloatArrayAccessor):
    KEY = "latitudes"

    def __init__(self):
        super().__init__()

    def get(self, handle, dtype=None):
        dtype = self.to_numpy_dtype(dtype)
        return super().get(handle, dtype=dtype)


class GribCodesLongitudeAccessor(GribCodesFloatArrayAccessor):
    KEY = "longitudes"

    def __init__(self):
        super().__init__()

    def get(self, handle, dtype=None):
        dtype = self.to_numpy_dtype(dtype)
        return super().get(handle, dtype=dtype)


VALUE_ACCESSOR = GribCodesValueAccessor()
LATITUDE_ACCESSOR = GribCodesLatitudeAccessor()
LONGITUDE_ACCESSOR = GribCodesLongitudeAccessor()


class GribCodesMessagePositionIndex(CodesMessagePositionIndex):
    MAGIC = b"GRIB"

    # This does not belong here, should be in the C library
    def _get_message_positions_part(self, fd, part):
        assert part is not None
        assert len(part) == 2

        offset = part[0]
        end_pos = part[0] + part[1] if part[1] > 0 else -1

        if os.lseek(fd, offset, os.SEEK_SET) != offset:
            return

        while True:
            code = os.read(fd, 4)
            if len(code) < 4:
                break

            if code != self.MAGIC:
                offset = os.lseek(fd, offset + 1, os.SEEK_SET)
                continue

            length = self._get_bytes(fd, 3)
            edition = self._get_bytes(fd, 1)

            if edition == 1:
                if length & 0x800000:
                    sec1len = self._get_bytes(fd, 3)
                    os.lseek(fd, 4, os.SEEK_CUR)
                    flags = self._get_bytes(fd, 1)
                    os.lseek(fd, sec1len - 8, os.SEEK_CUR)

                    if flags & (1 << 7):
                        sec2len = self._get_bytes(fd, 3)
                        os.lseek(fd, sec2len - 3, os.SEEK_CUR)

                    if flags & (1 << 6):
                        sec3len = self._get_bytes(fd, 3)
                        os.lseek(fd, sec3len - 3, os.SEEK_CUR)

                    sec4len = self._get_bytes(fd, 3)

                    if sec4len < 120:
                        length &= 0x7FFFFF
                        length *= 120
                        length -= sec4len
                        length += 4

            if edition == 2:
                length = self._get_bytes(fd, 8)

            if end_pos > 0 and offset + length > end_pos:
                return

            yield offset, length
            offset = os.lseek(fd, offset + length, os.SEEK_SET)


class GribCodesHandle(CodesHandle):
    PRODUCT_ID = eccodes.CODES_PRODUCT_GRIB

    # TODO: just a wrapper around the base class implementation to handle the
    # s,l,d qualifiers. Once these are implemented in the base class this method can
    # be removed. md5GridSection is also handled!
    def get(self, name, ktype=None, **kwargs):
        if name == "values":
            return self.get_values()
        elif name == "md5GridSection":
            return self.get_md5GridSection()
        elif name == "gridSpec":
            # Temporary measure because from ecCodes 2.41.0, when we ask for the gridSpec
            # ecCodes raises a gribapi.errors.FunctionNotImplementedError exception, which is
            # not caught by the  high level eccodes API even if a default value is provided.
            if "default" in kwargs:
                try:
                    return super().get(name, ktype, **kwargs)
                except Exception as e:
                    if "FunctionNotImplementedError" in str(type(e)):
                        return kwargs.get("default", None)

        return super().get(name, ktype, **kwargs)

    def get_md5GridSection(self):
        # Special case because:
        #
        # 1) eccodes is returning size > 1 for 'md5GridSection'
        # (size = 16 : it is the number of bytes of the value)
        # This is already fixed in eccodes 2.27.1
        #
        # 2) sometimes (see below), the value for "shapeOfTheEarth" is inconsistent.
        # This impacts the (computed on-the-fly) value of "md5GridSection".
        # ----------------
        # Example of data with inconsistent values:
        # S2S data, origin='ecmf', param='tp', step='24', number='0', date=['20201203','20200702']
        # the 'md5GridSection' are different
        # This is because one has "shapeOfTheEarth" set to 0, the other to 6.
        # This is only impacting the metadata.
        # Since this has no impact on the data itself,
        # this is unlikely to be fixed. Therefore this hacky patch.
        #
        # Obviously, the patch causes an inconsistency between the value of md5GridSection
        # read by this code, and the value read by another code without this patch.

        save = eccodes.codes_get_long(self._handle, "shapeOfTheEarth")
        eccodes.codes_set_long(self._handle, "shapeOfTheEarth", 255)
        result = eccodes.codes_get_string(self._handle, "md5GridSection")
        eccodes.codes_set_long(self._handle, "shapeOfTheEarth", save)
        return result

    def as_namespace(self, namespace, param="shortName"):
        r = {}
        ignore = {
            "distinctLatitudes",
            "distinctLongitudes",
            "distinctLatitudes",
            "latLonValues",
            "latitudes",
            "longitudes",
            "values",
            "bitmap",
        }
        for key in self.keys(namespace=namespace):
            if key not in ignore:
                r[key] = self.get(param if key == "param" else key)
        return r

    # TODO: once missing value handling is implemented in the base class this method
    # can be removed
    def get_values(self, dtype=None):
        eccodes.codes_set(self._handle, "missingValue", CodesHandle.MISSING_VALUE)
        vals = VALUE_ACCESSOR.get(self._handle, dtype=dtype)
        if self.get_long("bitmapPresent"):
            vals[vals == CodesHandle.MISSING_VALUE] = np.nan
        return vals

    def get_latitudes(self, dtype=None):
        return LATITUDE_ACCESSOR.get(self._handle, dtype=dtype)

    def get_longitudes(self, dtype=None):
        return LONGITUDE_ACCESSOR.get(self._handle, dtype=dtype)

    def get_data_points(self):
        return eccodes.codes_grib_get_data(self._handle)

    def set_values(self, values):
        try:
            assert self.path is None, "Only cloned handles can have values changed"
            eccodes.codes_set_values(self._handle, values.flatten())
        except Exception as e:
            LOG.error("Error setting values")
            LOG.exception(e)
            raise


class GribCodesReader(CodesReader):
    PRODUCT_ID = eccodes.CODES_PRODUCT_GRIB
    HANDLE_TYPE = GribCodesHandle


class GribField(Field):
    r"""Represent a GRIB message in a GRIB file.

    Parameters
    ----------
    path: str
        Path to the GRIB file
    offset: number
        File offset of the message (in bytes)
    length: number
        Size of the message (in bytes)
    """

    _handle = None

    def __init__(self, path, offset, length, handle_manager=None, use_metadata_cache=False):
        super().__init__()
        self.path = path
        self._offset = offset
        self._length = length
        self._handle_manager = handle_manager
        self._use_metadata_cache = use_metadata_cache

    @property
    def handle(self):
        r""":class:`CodesHandle`: Get an object providing access to the low level GRIB message structure."""
        if self._handle_manager is not None:
            handle = self._handle_manager.handle(self, self._create_handle)
            if handle is None:
                raise RuntimeError(f"Could not get a handle for offset={self.offset} in {self.path}")
            return handle

        # create a new handle and keep it in the field
        if self._handle is None:
            assert self._offset is not None
            self._handle = self._create_handle()
        return self._handle

    def _create_handle(self):
        return GribCodesReader.from_cache(self.path).at_offset(self.offset)

    def _values(self, dtype=None):
        return self.handle.get_values(dtype=dtype)

    @property
    def offset(self):
        r"""number: Gets the offset (in bytes) of the GRIB field within the GRIB file."""
        if self._offset is None:
            self._offset = int(self.handle.get("offset"))
        return self._offset

    @cached_property
    def _metadata(self):
        cache = self._use_metadata_cache
        if cache:
            cache = self._make_metadata_cache()
        return GribFieldMetadata(self, cache=cache)

    def _make_metadata_cache(self):
        return dict()

    def __repr__(self):
        return "GribField(%s,%s,%s,%s,%s,%s)" % (
            self._metadata.get("shortName", None),
            self._metadata.get("levelist", None),
            self._metadata.get("date", None),
            self._metadata.get("time", None),
            self._metadata.get("step", None),
            self._metadata.get("number", None),
        )

    # def write(self, f, **kwargs):
    #     r"""Write the message to a file object.

    #     Parameters
    #     ----------
    #     f: file object
    #         The target file object.
    #     bits_per_value: int or None
    #         Set the ``bitsPerValue`` GRIB key in the generated GRIB message. When
    #         None the ``bitsPerValue`` stored in the metadata will be used.
    #     """
    #     self.handle.write(f)
    #     # from earthkit.data.writers import write

    #     # write(f, self, **kwargs)

    def message(self):
        r"""Return a buffer containing the encoded message.

        Returns
        -------
        bytes
        """
        return self.handle.get_buffer()

    def clone(self, **kwargs):
        return ClonedGribField(self, **kwargs)

    def __getstate__(self):
        state = super().__getstate__()
        state["path"] = self.path
        state["offset"] = self._offset
        state["length"] = self._length
        state["use_metadata_cache"] = self._use_metadata_cache
        return state

    def __setstate__(self, state):
        self.path = state["path"]
        self._offset = state["offset"]
        self._length = state["length"]
        self._use_metadata_cache = state["use_metadata_cache"]
        self._handle_manager = None


class ClonedGribField(ClonedFieldCore, GribField):
    def __init__(self, field, **kwargs):
        ClonedFieldCore.__init__(self, field, **kwargs)
        self._handle = field._handle
        GribField.__init__(
            self,
            field.path,
            field._offset,
            field._length,
            handle_manager=field._handle_manager,
            use_metadata_cache=field._use_metadata_cache,
        )



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import os
import sys

import pytest

here = os.path.dirname(__file__)
sys.path.insert(0, here)

from lod_fixtures import build_lod_fieldlist  # noqa: E402


@pytest.mark.parametrize("mode", ["list-of-dicts", "loop"])
def test_lod_metadata_core(lod_ll_flat, mode):
    ds = build_lod_fieldlist(lod_ll_flat, mode)

    ref = [500, 850, 500, 850, 850, 600]
    assert ds.metadata("level") == ref

    assert ds[0].metadata("level") == 500
    assert ds[0].metadata("levelist") == 500

    assert ds[0].metadata("param") == "t"
    assert ds[0].metadata("shortName") == "t"

    dt_ref = datetime.datetime.fromisoformat("2018-08-01T09:00:00")
    assert ds[0].metadata("valid_datetime") == "2018-08-01T09:00:00"
    assert ds[0].metadata().valid_datetime() == dt_ref
    assert ds[0].datetime() == {"base_time": None, "valid_time": dt_ref}


@pytest.mark.parametrize("data", ["lod_ll_forecast_1", "lod_ll_forecast_2"])
@pytest.mark.parametrize("mode", ["list-of-dicts", "loop"])
def test_lod_metadata_forecast_a(request, data, mode):
    ds = build_lod_fieldlist(request.getfixturevalue(data), mode)

    valid_ref = datetime.datetime.fromisoformat("2018-08-01T09:00:00")
    base_ref = datetime.datetime.fromisoformat("2018-08-01T03:00:00")
    assert ds[0].metadata("valid_datetime") == "2018-08-01T09:00:00"
    assert ds[0].metadata().valid_datetime() == valid_ref
    assert ds[0].metadata("base_datetime") == "2018-08-01T03:00:00"
    assert ds[0].metadata().base_datetime() == base_ref
    assert ds[0].datetime() == {"base_time": base_ref, "valid_time": valid_ref}
    assert ds[0].metadata("step") == 6
    assert ds[0].metadata("step_timedelta") == datetime.timedelta(hours=6)
    assert ds[0].metadata().step_timedelta() == datetime.timedelta(hours=6)
    assert ds[0].metadata("forecast_reference_time") == "2018-08-01T03:00:00"

    assert ds[1].metadata("forecast_reference_time") == "2018-08-01T03:00:00"


@pytest.mark.parametrize("data", ["lod_ll_forecast_3", "lod_ll_forecast_4"])
@pytest.mark.parametrize("mode", ["list-of-dicts", "loop"])
def test_lod_metadata_forecast_b(request, data, mode):
    ds = build_lod_fieldlist(request.getfixturevalue(data), mode)

    valid_ref = datetime.datetime.fromisoformat("2018-08-01T09:00:00")
    base_ref = datetime.datetime.fromisoformat("2018-08-01T03:00:00")
    assert ds[0].metadata("valid_datetime") == "2018-08-01T09:00:00"
    assert ds[0].metadata().valid_datetime() == valid_ref
    assert ds[0].metadata("base_datetime") == "2018-08-01T03:00:00"
    assert ds[0].metadata().base_datetime() == base_ref
    assert ds[0].datetime() == {"base_time": base_ref, "valid_time": valid_ref}
    assert ds[0].metadata("step") == datetime.timedelta(hours=6)
    assert ds[0].metadata("step_timedelta") == datetime.timedelta(hours=6)
    assert ds[0].metadata().step_timedelta() == datetime.timedelta(hours=6)
    assert ds[0].metadata("forecast_reference_time") == "2018-08-01T03:00:00"

    assert ds[1].metadata("forecast_reference_time") == "2018-08-01T03:00:00"



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from functools import cached_property

from earthkit.data.core.fieldlist import Field
from earthkit.data.core.metadata import WrappedMetadata
from earthkit.data.utils.dates import date_to_grib
from earthkit.data.utils.dates import datetime_from_grib
from earthkit.data.utils.dates import time_to_grib
from earthkit.data.utils.dates import to_timedelta

from .index import GribFieldList

LOG = logging.getLogger(__name__)


class VirtualGribField(Field):
    def __init__(self, owner, request, metadata_alias, reference=None):
        self.owner = owner
        self.request = request
        self.metadata_alias = metadata_alias
        self.reference = reference

        self.extra = {}
        if "param" in request:
            if "shortName" not in request:
                self.extra = self.owner._get_info(self.request["param"])
            if "shortName" not in self.extra:
                self.extra["shortName"] = request["param"]
            else:
                self.extra["param"] = self.extra["shortName"]

    @property
    def _data_datetime(self):
        if "date" in self.request and "time" in self.request:
            return datetime_from_grib(self.request["date"], self.request["time"])
        return None

    @property
    def _valid_datetime(self):
        base = self._data_datetime
        if base is None:
            return None
        if "step" in self.request:
            step = self.request["step"]
            return base + to_timedelta(step)
        return base

    def _attributes(self, names, remapping=None, joiner=None, default=None):
        result = {}
        metadata = self.metadata
        if remapping is not None:
            metadata = remapping(metadata, joiner=joiner)

        for name in names:
            result[name] = metadata(name, default=default)
        return result

    def metadata(self, *keys, astype=None, remapping=None, patches=None, **kwargs):
        if (not kwargs or kwargs == {"default": None}) and keys:
            if isinstance(keys[0], (list, tuple)):
                keys = keys[0]
            if keys and isinstance(keys[0], str):
                r = []
                for k in keys:
                    r.append(self._one_metadata(k, astype=astype, remapping=remapping, patches=patches))
                if len(r) == 1:
                    return r[0]
                return r

        return super().metadata(
            *keys,
            astype=astype,
            remapping=remapping,
            patches=patches,
            **kwargs,
        )

    def _one_metadata(self, key, remapping=None, patches=None, **kwargs):
        # print(f"one_metadata key={key} kwargs={kwargs}")
        if key in self.extra:
            return self.extra[key]
        if key in self.request:
            return self.request[key]
        if key in self.metadata_alias and key in self.request:
            return self.request[self.metadata_alias[key]]

        if key == "number":
            return 0
        if key == "validityDate":
            return date_to_grib(self._valid_datetime)
        if key == "validityTime":
            return time_to_grib(self._valid_datetime)
        if key in ("forecast_reference_time", "base_time", "base_datetime"):
            # print("here")
            return self._data_datetime.isoformat()
        if key in ("valid_datetime", "valid_time"):
            return self._valid_datetime.isoformat()

        return self._metadata.get(key, **kwargs)

    @property
    def _metadata(self):
        r = {**self.request, **self.extra}
        for k, v in self.metadata_alias.items():
            if k not in r and v in r:
                r[k] = r[v]

        return WrappedMetadata(self.owner.reference._metadata, extra=r)

    def _values(self, dtype=None):
        return self._field._values(dtype=dtype)

    @property
    def _field(self):
        if self.reference:
            return self.reference
        else:
            return self.owner.retriever.get(self.request)[0]


class VirtualGribFieldList(GribFieldList):
    def __init__(self, request_mapper, retriever):
        self.request_mapper = request_mapper
        self.retriever = retriever

        self._info_cache = {}

    def __len__(self):
        return len(self.request_mapper)

    def mutate(self):
        return self

    @cached_property
    def reference(self):
        return self.retriever.get(self.request_mapper.request_at(0))[0]

    def _getitem(self, n):
        if isinstance(n, int):
            if n < 0:
                n += len(self)
            if n >= len(self):
                raise IndexError(f"Index {n} out of range")

            return VirtualGribField(
                self,
                self.request_mapper.request_at(n),
                self.request_mapper.metadata_alias,
                reference=self.reference if n == 0 else None,
            )

    def _get_info(self, param):
        if param in self._info_cache:
            return self._info_cache[param]

        ref_request = self.request_mapper.request_at(0)
        if param == ref_request.get("param"):
            r = self.reference._attributes(["shortName", "name", "units", "cfName"])
        else:
            md = self.reference.metadata().override(paramId=param)
            r = {k: md.get(k, None) for k in ["shortName", "name", "units", "cfName"]}

        self._info_cache[param] = r
        return r



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.readers.netcdf.field import XArrayField
from earthkit.data.testing import NO_ZARR
from earthkit.data.testing import earthkit_test_data_file


@pytest.mark.skipif(NO_ZARR, reason="Zarr not installed")
def test_zarr_source():
    ds = from_source("zarr", earthkit_test_data_file("test_zarr/"))
    assert len(ds) == 4
    assert isinstance(ds[0], XArrayField)
    assert isinstance(ds[1], XArrayField)
    assert isinstance(ds[2], XArrayField)
    assert isinstance(ds[3], XArrayField)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging
import math

LOG = logging.getLogger(__name__)


class DictDiffInfo:
    def __init__(self, same, diff_dict={}, diff_text=str()):
        self.same = same
        self.diff_dict = dict(**diff_dict)
        self.diff_text = diff_text


class DictDiff:
    @staticmethod
    def diff(vals1, vals2):
        if not isinstance(vals1, dict):
            raise ValueError(f"Unsupported type for vals1: {type(vals1)}. Expecting dict")

        if not isinstance(vals2, dict):
            raise ValueError(f"Unsupported type for vals2: {type(vals2)}. Expecting dict")

        diff_dict = {}
        if len(vals1) != len(vals2):
            return ListDiffInfo(False, ListDiff.VALUE_DIFF, f"Length mismatch: {len(vals1)} != {len(vals2)}")

        for k, v1 in vals1.items():
            if k not in vals2:
                diff_dict[k] = (None, v1)
            else:
                v2 = vals2[k]
                same, _ = ListDiff._compare(v1, v2)
                if not same:
                    diff_dict[k] = (v2, v1)

        if diff_dict:
            diff_text = ", ".join([f"{k}: {v[0]} != {v[1]}" for k, v in diff_dict.items()])
            return DictDiffInfo(False, diff_dict=diff_dict, diff_text=diff_text)
        else:
            return DictDiffInfo(True)


class ListDiffInfo:
    def __init__(self, same, diff_type=None, diff_text=str(), diff_index=-1):
        self.same = same
        self.type = diff_type
        self.diff_text = diff_text
        self.diff_index = diff_index


class ListDiff:
    VALUE_DIFF = 0
    TYPE_DIFF = 1

    @staticmethod
    def _compare(v1, v2):
        if isinstance(v1, int) and isinstance(v2, int):
            return v1 == v2, ListDiff.VALUE_DIFF
        elif isinstance(v1, float) and isinstance(v2, float):
            return math.isclose(v1, v2, rel_tol=1e-9), ListDiff.VALUE_DIFF
        elif isinstance(v1, str) and isinstance(v2, str):
            return v1 == v2, ListDiff.VALUE_DIFF
        elif isinstance(v1, datetime.timedelta) and isinstance(v2, datetime.timedelta):
            return v1 == v2, ListDiff.VALUE_DIFF
        elif type(v1) is not type(v2):
            return False, ListDiff.TYPE_DIFF
        else:
            raise ValueError(f"Unsupported type: {type(v1)}")

    @staticmethod
    def diff(vals1, vals2, name=str()):
        if not isinstance(vals1, (list, tuple)):
            raise ValueError(f"Unsupported type for vals1: {type(vals1)}. Expecting list/tuple")

        if not isinstance(vals2, (list, tuple)):
            raise ValueError(f"Unsupported type for vals2: {type(vals2)}. Expecting list/tuple")

        if len(vals1) != len(vals2):
            return ListDiffInfo(False, ListDiff.VALUE_DIFF, f"Length mismatch: {len(vals1)} != {len(vals2)}")

        for i, (v1, v2) in enumerate(zip(vals1, vals2)):
            same, diff = ListDiff._compare(v1, v2)
            if not same:
                if diff == ListDiff.VALUE_DIFF:
                    diff = f"Value mismatch at {name}[{i}]: {v1} != {v2}"
                elif diff == ListDiff.TYPE_DIFF:
                    diff = f"Type mismatch at {name}[{i}]: {type(v1)} != {type(v2)}"
                return ListDiffInfo(False, ListDiff.VALUE_DIFF, diff, i)
        return ListDiffInfo(True)


def check_coords(self, var_name, coord_name, coords, ref_coords):
    if coord_name in ref_coords:
        # from .diff import ListDiff
        from .diff import list_to_str

        v1 = ref_coords[coord_name].vals
        v2 = coords[coord_name]
        diff = ListDiff.diff(v1, v2, name=coord_name)
        if not diff.same:
            raise ValueError(
                (
                    f'Variable "{var_name}" has inconsistent dimension "{coord_name}" compared '
                    f"to other variables. Expected values: {list_to_str(v1)}, "
                    f"got: {list_to_str(v2)}. {diff.diff_text}"
                )
            )



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from pprint import pformat

import yaml

from earthkit.data.utils.xarray.profile import Profile


def format_table_cell(v, indent=13):
    def _indent(text, indent_first=False):
        lines = text.split("\n")
        r = ""
        if len(lines) == 1:
            if indent_first:
                r += " " * indent
            return r + lines[0]
        elif len(lines) > 1:
            r = ""
            for line in lines:
                if line:
                    if r or indent_first:
                        r += " " * indent
                    r += "| " + line + "\n"
            if r.endswith("\n"):
                r = r[:-1]
            return r
        return v

    if isinstance(v, list):
        return str(v)
    return _indent(pformat(v, sort_dicts=False, indent=1))


def make_table(d):
    t = """
   .. tab:: Table

      .. list-table::
         :header-rows: 1
         :widths: 20 80

         * - Name
           - Value"""

    for k, v in d.items():
        if k != "coord_attrs":
            t += f"""
         * - {k}
           - {format_table_cell(v, indent=13)}"""

    if "coord_attrs" in d:
        t += """

      Please note that the following options cannot be specified as a kwarg to :func:`to_xarray`.

      .. list-table::
         :header-rows: 1
         :widths: 20 80

         * - Name
           - Value"""

        for k in ["coord_attrs"]:
            t += f"""
         * - {k}
           - {format_table_cell(d[k], indent=13)}"""

    return t


def make_dict(d):
    d = pformat(d, sort_dicts=False, indent=1)
    t = """

   .. tab:: Dictionary

      .. code-block:: python

    """
    for line in d.split("\n"):
        t += f"         {line}\n"

    return t


def make_yaml(d):
    d = yaml.dump(d, default_flow_style=False, sort_keys=False)
    t = """

   .. tab:: YAML

      .. code-block:: yaml

    """
    for line in d.split("\n"):
        t += f"         {line}\n"

    return t


def execute(*args):
    name = args[0]
    if name == "__py_None__":
        name = None
    d = Profile.to_docs(name)

    t = f"""
.. tabs::

{make_table(d)}

{make_dict(d)}

{make_yaml(d)}

"""
    print(t)


if __name__ == "__main__":
    execute()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

from earthkit.data import from_source
from earthkit.data.core.fieldlist import FieldList
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import get_array_backend


def load_array_fl(num, array_backend=None):
    assert num in [1, 2, 3]
    files = ["test.grib", "test6.grib", "tuv_pl.grib"]
    files = files[:num]

    ds_in = []
    md = []
    for fname in files:
        ds = from_source("file", earthkit_examples_file(fname))
        ds_in.append(ds.to_fieldlist(array_backend=array_backend))
        md += ds_in[-1].metadata("param")

    ds = []
    for x in ds_in:
        # print("len", len(x))
        ds.append(FieldList.from_array(x.values, [m.override(edition=1) for m in x.metadata()]))

    return (*ds, md)


def load_array_fl_file(fname, array_backend=None):
    ds_in = from_source("file", earthkit_examples_file(fname))
    ds_in = ds_in.to_fieldlist(array_backend=array_backend)
    md = ds_in.metadata("param")

    ds = FieldList.from_array(ds_in.values, [m.override(edition=1) for m in ds_in.metadata()])

    return (ds, md)


def check_array_fl(ds, ds_input, md_full, array_backend=None):
    assert len(ds_input) in [1, 2, 3]

    array_backend = get_array_backend(array_backend)

    assert len(ds) == len(md_full)
    assert ds.metadata("param") == md_full
    assert array_backend.allclose(ds[0].values, ds_input[0][0].values)

    # # values metadata
    # keys = ["min", "max"]
    # for k in keys:
    #     assert np.isclose(ds[0].metadata(k), ds_input[0][0].metadata(k))

    # check slice
    r = ds[1]
    assert r.metadata("param") == "msl"

    if len(ds_input) > 1:
        r = ds[1:3]
        assert len(r) == 2
        assert r.metadata("param") == ["msl", "t"]
        assert r[0].metadata("param") == "msl"
        assert r[1].metadata("param") == "t"
        assert array_backend.allclose(r[0].values, ds_input[0][1].values)
        assert array_backend.allclose(r[1].values, ds_input[1][0].values)

        # check sel
        r = ds.sel(shortName="msl")
        assert len(r) == 1
        assert r.metadata("shortName") == ["msl"]
        assert r[0].metadata("param") == "msl"
        assert array_backend.allclose(r[0].values, ds_input[0][1].values)

    if len(ds_input) == 3:
        r = ds[1:13:4]
        assert len(r) == 3
        assert r.metadata("param") == ["msl", "t", "u"]
        assert r[0].metadata("param") == "msl"
        assert r[1].metadata("param") == "t"
        assert r[2].metadata("param") == "u"


def check_array_fl_from_to_fieldlist(ds, ds_input, md_full, array_backend=None, flatten=False, dtype=None):
    assert len(ds_input) in [1, 2, 3]
    assert len(ds) == len(md_full)
    assert ds.metadata("param") == md_full

    array_backend = get_array_backend(array_backend)

    np_kwargs = {"flatten": flatten, "dtype": dtype}

    assert array_backend.allclose(ds[0].to_array(**np_kwargs), ds_input[0][0].to_array(**np_kwargs))

    assert ds.to_array(**np_kwargs).shape == ds_input[0].to_array(**np_kwargs).shape

    # check slice
    r = ds[1]
    assert r.metadata("param") == "msl"

    if len(ds_input) > 1:
        r = ds[1:3]
        assert len(r) == 2
        assert r.metadata("param") == ["msl", "t"]
        assert r[0].metadata("param") == "msl"
        assert r[1].metadata("param") == "t"
        assert array_backend.allclose(r[0].to_array(**np_kwargs), ds_input[0][1].to_array(**np_kwargs))
        assert array_backend.allclose(r[1].to_array(**np_kwargs), ds_input[1][0].to_array(**np_kwargs))

        # check sel
        r = ds.sel(shortName="msl")
        assert len(r) == 1
        assert r.metadata("shortName") == ["msl"]
        assert r[0].metadata("param") == "msl"
        assert array_backend.allclose(r[0].to_array(**np_kwargs), ds_input[0][1].to_array(**np_kwargs))

    if len(ds_input) == 3:
        r = ds[1:13:4]
        assert len(r) == 3
        assert r.metadata("param") == ["msl", "t", "u"]
        assert r[0].metadata("param") == "msl"
        assert r[1].metadata("param") == "t"
        assert r[2].metadata("param") == "u"


def check_save_to_disk(ds, len_ref, meta_ref):
    tmp = temp_file()
    ds.to_target("file", tmp.path)
    assert os.path.exists(tmp.path)
    r_tmp = from_source("file", tmp.path)
    assert len(r_tmp) == len_ref
    assert r_tmp.metadata("shortName") == meta_ref
    r_tmp = None



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import re
from abc import ABCMeta
from abc import abstractmethod
from functools import lru_cache

from earthkit.data.core.gridspec import GridSpec
from earthkit.data.core.metadata import RawMetadata

LOG = logging.getLogger(__name__)

# NOTE: this is a temporary code until the full gridspec
# implementation is available via earthkit-geo.
#
# Limitations:
# - The gridspec will only be generated for a few unrotated grids.
# - The gridspec to metadata conversion is only supported for
#   regular_ll grids.


FULL_GLOBE = 360.0
FULL_GLOBE_EPS = 1e-7

REGULAR_GG_PATTERN = re.compile(r"[OoNn]\d+")
REDUCED_GG_PATTERN = re.compile(r"[Ff]\d+|\d+")


def make_gridspec(metadata):
    maker = GridSpecMaker(metadata)
    return GridSpec(maker.make())


class GridSpecConf:
    def __init__(self):
        self._config = None
        self._grid_types = None
        self._schema = None

    def _load(self):
        if self._config is None:
            import yaml

            from earthkit.data.utils.paths import earthkit_conf_file

            # # schema
            # with open(earthkit_conf_file("gridspec_schema.json"), "r") as f:
            #     self._schema = json.load(f)
            # gridspec config
            with open(earthkit_conf_file("gridspec.yaml"), "r") as f:
                self._config = yaml.safe_load(f)

            # add gridspec key to grib key mapping to conf
            d = {}
            for k, v in self._config["grib_key_map"].items():
                if v in d:
                    k_act = d[v]
                    if isinstance(k_act, tuple):
                        k = tuple([*k_act, k])
                    else:
                        k = tuple([k_act, k])
                d[v] = k
            self._config["spec_key_map"] = d

            # assign conf to GRIB gridType
            self._grid_types = {}
            for k, v in self._config["types"].items():
                g = v["grid_type"]
                self._grid_types[g] = k
                g = v.get("rotated_type", None)
                if g is not None:
                    self._grid_types[g] = k

    @property
    def config(self):
        self._load()
        return self._config

    @property
    def grid_types(self):
        self._load()
        return self._grid_types

    def remap_keys_to_grib(self, spec):
        self._load()
        spec_to_grib = self._config["spec_key_map"]
        r = {}
        for k, v in spec.items():
            grib_key = spec_to_grib[k]
            if isinstance(grib_key, tuple):
                for x in grib_key:
                    r[x] = v
            else:
                r[grib_key] = v
        return r

    def validate(self, gridspec):
        pass
        # self._load()
        # from jsonschema import validate

        # validate(instance=gridspec, schema=self._schema)


CONF = GridSpecConf()


class GridSpecMaker(RawMetadata):
    POSITIVE_SCAN_DIR = 1
    NEGATIVE_SCAN_DIR = -1

    def __init__(self, metadata):
        self.conf = CONF.config

        # remap metadata keys and get values
        d = {}
        for k, v in self.conf["grib_key_map"].items():
            act_val = d.get(v, None)
            if act_val is None:
                d[v] = metadata.get(k, None)

        # determine grid type
        grid_type = d["grid_type"]
        self.grid_type = CONF.grid_types.get(grid_type, None)
        if self.grid_type is None:
            raise ValueError(f"Unsupported grib grid type={grid_type}")

        if "rotated" in d["grid_type"]:
            raise ValueError(f"gridspec is not supported for rotated grids {grid_type=}")
        self.grid_conf = dict(self.conf["types"][self.grid_type])
        d["type"] = self.grid_type

        self.getters = {
            "N": self.N,
            "area": self.area,
            "H": self.H,
        }

        super().__init__(d)

    def make(self):
        d = {}

        for v in self.grid_conf["spec"]:
            self._add_key_to_spec(v, d)

        if "rotated" not in self["grid_type"]:
            for k in self.conf["rotation_keys"]:
                d.pop(k, None)

        CONF.validate(d)
        return d

    def _add_key_to_spec(self, item, d):
        if isinstance(item, str):
            key = item
            method = self.getters.get(key, self.get)
            d[key] = method(key)
        elif isinstance(item, dict):
            for k, v in item.items():
                method = self.getters.get(k, self.get_list)
                r = method(v)
                d[k] = r[0] if len(r) == 1 else r
        elif isinstance(item, list):
            for v in item:
                self._add_key_to_spec(v, d)
        else:
            raise TypeError(f"Unsupported item type={type(item)}")

    def get_list(self, item):
        r = []
        for k in item:
            method = self.getters.get(k, None)
            if method is not None:
                v = method()
            else:
                v = self.get(k)
            r.append(v)
        return r

    def area(self, item):
        a = {}
        a["north"] = max(self["first_lat"], self["last_lat"])
        a["south"] = min(self["first_lat"], self["last_lat"])
        a["west"] = self.west()
        a["east"] = self.east()
        return [a[k] for k in item]

    def west(self):
        if self.x_scan_dir() == self.POSITIVE_SCAN_DIR:
            return self["first_lon"]
        else:
            return self["last_lon"]

    def east(self):
        if self.x_scan_dir() == self.POSITIVE_SCAN_DIR:
            return self["last_lon"]
        else:
            return self["first_lon"]

    def x_scan_dir(self):
        v = self.get("i_scans_negatively", None)
        if v is not None:
            return self.POSITIVE_SCAN_DIR if v == 0 else self.NEGATIVE_SCAN_DIR
        else:
            raise ValueError("Could not determine i-direction scanning mode")

    def N(self):
        label = self.grid_conf["N_label"]
        if isinstance(label, dict):
            if "octahedral" not in label:
                raise ValueError(f"octahedral missing from N label config={label}")
            octahedral = 1 if self.get("octahedral", 0) == 1 else 0
            label = label["octahedral"][octahedral]
        elif not isinstance(label, str):
            raise ValueError(f"invalid N label config={label}")
        return label + str(self["N"])

    def H(self):
        return f"H{self['n_side']}"


class GridSpecConverter(metaclass=ABCMeta):
    SPEC_GRID_TYPE = None

    def __init__(self, spec, spec_type, edition):
        self.spec = spec
        self.spec_type = spec_type
        self.conf = CONF.config["types"][spec_type]
        self.edition = edition
        self.grid_size = 0

    def run(self):
        # the order might matter
        d = self.add_grid_type()
        d.update(self.add_grid())
        d.update(self.add_rotation())
        d.update(self.add_scanning())
        d = CONF.remap_keys_to_grib(d)
        return d

    @staticmethod
    def to_metadata(spec, edition=2):
        CONF.validate(spec)

        if "rotation" in spec:
            raise ValueError(
                (
                    "GridSpecConverter: grispec cannot contain the 'rotation' keyword. "
                    "Only unrotated grids are supported!"
                )
            )

        spec_type, maker = GridSpecConverter.infer_spec_type(spec)

        # create converter and generate metadata
        # maker = gridspec_converters.get(spec_type, None)
        if maker is None:
            raise ValueError(f"GridSpecConverter: unsupported gridspec type={spec_type}")
        else:
            converter = maker(spec, spec_type, edition)
            return converter.run(), converter.grid_size

    @staticmethod
    def infer_spec_type(spec):
        spec_type = spec.get("type", None)
        if spec_type is None:
            if "grid" not in spec:
                raise ValueError(f"GridSpecConverter: unsupported gridspec={spec}")
            grid = spec["grid"]
            for k, gs in gridspec_converters.items():
                if gs.type_match(grid):
                    return k, gs

        if spec_type is None:
            raise ValueError(f"GridSpecConverter: could not determine type of gridspec={spec}")

        return spec_type, gridspec_converters.get(spec_type, None)

    # @staticmethod
    # def infer_spec_type(spec):
    #     spec_type = spec.get("type", None)
    #     # when no type specified the grid must be regular_ll or gaussian
    #     if spec_type is None:
    #         grid = spec["grid"]
    #         # regular_ll: the grid is in the form of [dx, dy]
    #         if isinstance(grid, list) and len(grid) == 2:
    #             spec_type = "regular_ll"
    #         # gaussian: the grid=N as a str or int
    #         elif isinstance(grid, (str, int)):
    #             spec_type = GridSpecConverter.infer_gaussian_type(grid)

    #     if spec_type is None:
    #         raise ValueError(f"Could not determine type of gridspec={spec}")

    #     return spec_type

    # @staticmethod
    # def infer_gaussian_type(grid):
    #     """Determine gridspec type for Gaussian grids"""
    #     grid_type = ""
    #     if isinstance(grid, str) and len(grid) > 0:
    #         try:
    #             if grid[0] == "F":
    #                 grid_type = "regular_gg"
    #             elif grid[0] in ["N", "O"]:
    #                 grid_type = "reduced_gg"
    #             else:
    #                 grid_type = "regular_gg"
    #                 _ = int(grid)
    #         except Exception:
    #             raise ValueError(f"Invalid Gaussian grid description str={grid}")
    #     elif isinstance(grid, int):
    #         grid_type = "regular_gg"
    #     else:
    #         raise ValueError(f"Invalid Gaussian grid description={grid}")

    #     return grid_type

    def add_grid_type(self):
        d = {}
        d["grid_type"] = self.conf["grid_type"]

        # rotation = self.add_rotation()
        # if rotation:
        #     rotated_type = self.conf.get("rotated_type", None)
        #     if rotated_type is None:
        #         raise ValueError(
        #             f"GridSpecConverter: rotation is not supported for gridspec type={self.spec_type}"
        #         )
        #     d["grid_type"] = rotated_type
        #     d.update(rotation)

        return d

    @abstractmethod
    def add_grid(self):
        pass

    def add_rotation(self):
        return dict()
        # d = {}
        # rotation = self.spec.get("rotation", None)
        # if rotation is not None:

        #     if not isinstance(rotation, list) or len(rotation) != 2:
        #         raise ValueError(f"Invalid rotation in grid spec={rotation}")
        #     d["lat_south_pole"] = rotation[0]
        #     d["lon_south_pole"] = rotation[1]
        #     d["angle_of_rotation"] = self.get("angle_of_rotation")
        # return d

    def add_scanning(self):
        d = {}
        keys = {
            "j_points_consecutive": 0,
            "i_scans_negatively": 0,
            "j_scans_positively": 0,
        }
        for k, v in keys.items():
            d[k] = self.get(k, default=v, transform=self.to_zero_one)
        return d

    def _parse_scanning(self):
        d = self.add_scanning()
        return (d["i_scans_negatively"], d["j_scans_positively"])

    def to_zero_one(self, v):
        return 1 if (v == 1 or v is True) else 0

    def get(self, key, default=None, transform=None):
        v = self.spec.get(key, default)
        if v is not None and callable(transform):
            return transform(v)
        return v

    @staticmethod
    def normalise_lon(lon, minimum):
        while lon < minimum:
            lon += FULL_GLOBE
        while lon >= minimum + FULL_GLOBE:
            lon -= FULL_GLOBE
        return lon


class LatLonGridSpecConverter(GridSpecConverter):
    SPEC_GRID_TYPE = "regular_ll"

    @staticmethod
    @lru_cache
    def global_grids():
        import json

        from earthkit.data.utils.paths import earthkit_conf_file

        with open(earthkit_conf_file("global_grids.json"), "r") as f:
            return json.load(f)

    def _parse_ew(self, dx, west, east):
        nx = self.spec.get("nx", None)
        west = self.normalise_lon(west, 0)
        east = self.normalise_lon(east, 0)
        global_ew = False
        if east < west:
            east += FULL_GLOBE
        if abs(east - west) < FULL_GLOBE_EPS:
            east = west + FULL_GLOBE
            global_ew = True
        elif abs(east - west - FULL_GLOBE) < FULL_GLOBE_EPS:
            global_ew = True
        assert west >= 0
        assert east > west

        if global_ew:
            east -= abs(dx)

        if nx is None:
            d_lon = east - west
            nx = int(d_lon / dx) + 1
            eps = dx / 3
            if abs(abs((nx - 1) * dx) - abs(d_lon)) > eps:
                if abs((nx - 1) * dx) > abs(d_lon):
                    nx -= 1
                else:
                    nx += 1

        west = self.normalise_lon(west, 0)
        east = self.normalise_lon(east, 0)
        if self.edition == 1:
            if west > east:
                west = self.normalise_lon(west, -180)

        return nx, west, east

    def _parse_ns(self, dy, north, south):
        ny = self.spec.get("ny", None)
        if ny is None:
            d_lat = abs(north - south)
            ny = int(d_lat / dy) + 1
            eps = dy / 3
            if abs(abs((ny - 1) * dy) - abs(d_lat)) > eps:
                if abs((ny - 1) * dy) > abs(d_lat):
                    ny -= 1
                else:
                    ny += 1
        return ny, north, south

    def add_grid(self):
        d = {}
        dx, dy = self.spec.get("grid", [1, 1])

        area = self.spec.get("area", None)
        if isinstance(area, list) and len(area) == 4:
            north, west, south, east = area
            nx, west, east = self._parse_ew(dx, west, east)
            ny, north, south = self._parse_ns(dy, north, south)

            # apply i and j scanning directions
            i_scan_neg, j_scan_pos = self._parse_scanning()
            if j_scan_pos == 1:
                north, south = south, north
            if i_scan_neg == 1:
                west, east = east, west
        elif dx == dy:
            # supported global grid
            p = self.global_grids().get(str(dx), None)
            if p is not None:
                north, west, south, east = p["area"]
                ny, nx = p["shape"]
            else:
                raise ValueError("Unsupported gridspec={self.spec}")
        else:
            raise ValueError("Unsupported gridspec={self.spec}")

        d["nx"] = nx
        d["ny"] = ny
        d["dx"] = dx
        d["dy"] = dy
        d["first_lat"] = north
        d["last_lat"] = south
        d["first_lon"] = west
        d["last_lon"] = east

        self.grid_size = nx * ny

        return d

    @staticmethod
    def type_match(grid):
        if isinstance(grid, list) and len(grid) == 2:
            return True
        return False


class RegularGaussianGridSpecConverter(GridSpecConverter):
    SPEC_GRID_TYPE = "regular_gg"

    def add_grid(self):
        grid = self.spec.get("grid", None)
        if not RegularGaussianGridSpecConverter.type_match(grid):
            raise ValueError(f"Invalid {grid=}")
        if isinstance(grid, str):
            try:
                if grid[0] == "F":
                    N = int(grid[1:])
                else:
                    N = int(N)
            except Exception as e:
                raise ValueError(f"Invalid {grid=} {e}")
        elif isinstance(grid, int):
            N = grid
        else:
            raise ValueError(f"Invalid {grid=}")
        if N < 1 or N > 1000000:
            raise ValueError(f"Invalid {N=}")
        d = dict(N=N)
        return d

    @staticmethod
    def type_match(grid):
        if isinstance(grid, int):
            grid = str(grid)
        if isinstance(grid, str):
            if REGULAR_GG_PATTERN.match(grid):
                return True
        return False


class ReducedGaussianGridSpecConverter(GridSpecConverter):
    SPEC_GRID_TYPE = "reduced_gg"

    def add_grid(self):
        grid = self.spec.get("grid", None)
        octahedral = self.spec.get("octahedral", 0)
        if not isinstance(grid, str) or not REDUCED_GG_PATTERN.match(grid):
            raise ValueError(f"Invalid {grid=}")
        try:
            if grid[0] == "N":
                N = int(grid[1:])
                octahedral = 0
            elif grid[0] == "O":
                N = int(grid[1:])
                octahedral = 1
        except Exception as e:
            raise ValueError(f"Invalid {grid=} {e}")
        if N < 1 or N > 1000000:
            raise ValueError(f"Invalid {N=}")
        d = dict(N=N, octahedral=octahedral)
        return d

    @staticmethod
    def type_match(grid):
        if isinstance(grid, str):
            if REDUCED_GG_PATTERN.match(grid):
                return True
        return False


gridspec_converters = {}
for x in [
    LatLonGridSpecConverter,
    # RegularGaussianGridSpecConverter,
    # ReducedGaussianGridSpecConverter,
]:
    gridspec_converters[x.SPEC_GRID_TYPE] = x



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import NO_MARS
from earthkit.data.testing import NO_MARS_API
from earthkit.data.testing import NO_MARS_DIRECT
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import write_to_file


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_MARS, reason="No access to MARS")
@pytest.mark.parametrize("prompt", [True, False])
def test_mars_grib_1_prompt(prompt):
    s = from_source(
        "mars",
        param=["2t", "msl"],
        levtype="sfc",
        area=[50, -50, 20, 50],
        grid=[2, 2],
        prompt=prompt,
        date="2023-05-10",
    )
    assert len(s) == 2


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_MARS, reason="No access to MARS")
def test_mars_grib_2():
    s = from_source(
        "mars",
        param=["2t", "msl"],
        levtype="sfc",
        area=[50, -50, 20, 50],
        grid=[2, 2],
        date="2023-05-10",
        split_on="param",
    )
    assert len(s) == 2


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_MARS, reason="No access to MARS")
def test_mars_grib_expect_any_1():
    ds = from_source(
        "mars",
        expect="any",
        param=["2t", "msl"],
        levtype="sfc",
        area=[50, -50, 20, 50],
        grid=[2, 2],
        date="1054-05-10",
    )

    assert len(ds) == 0


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_MARS, reason="No access to MARS")
def test_mars_grib_expect_any_2():
    with pytest.raises(Exception):
        from_source(
            "mars",
            param=["2t", "msl"],
            levtype="sfc",
            area=[50, -50, 20, 50],
            grid=[2, 2],
            date="1054-05-10",
        )


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_MARS, reason="No access to MARS")
def test_mars_grib_log_1():
    s = from_source(
        "mars",
        prompt=False,
        log="default",
        param=["2t", "msl"],
        levtype="sfc",
        area=[50, -50, 20, 50],
        grid=[2, 2],
        date="2023-05-10",
    )
    assert len(s) == 2


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_MARS, reason="No access to MARS")
def test_mars_grib_log_2():
    s = from_source(
        "mars",
        prompt=False,
        log=None,
        param=["2t", "msl"],
        levtype="sfc",
        area=[50, -50, 20, 50],
        grid=[2, 2],
        date="2023-05-10",
    )
    assert len(s) == 2


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(not NO_MARS_DIRECT or NO_MARS_API, reason="No access to MARS")
def test_mars_grib_log_3():
    res = 0

    def _my_log(msg):
        nonlocal res
        res = 1

    s = from_source(
        "mars",
        prompt=False,
        log=_my_log,
        param=["2t", "msl"],
        levtype="sfc",
        area=[50, -50, 20, 50],
        grid=[2, 2],
        date="2023-05-10",
    )
    assert len(s) == 2


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_MARS_DIRECT and not NO_MARS_API, reason="No access to MARS")
def test_mars_grib_log_4():
    with temp_file() as tmp:
        with open(tmp, "w") as f:
            s = from_source(
                "mars",
                prompt=False,
                log={"stdout": f},
                param=["2t", "msl"],
                levtype="sfc",
                area=[50, -50, 20, 50],
                grid=[2, 2],
                date="2023-05-10",
            )
            assert len(s) == 2

        t = ""
        with open(tmp) as f:
            t = f.read()

        assert t


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_mars_grib_save(write_method):
    ds = from_source(
        "mars",
        param="2t",
        levtype="sfc",
        grid=[30, 30],
        date=-1,
    )
    assert len(ds) == 1

    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds)

        ds1 = from_source("file", tmp)
        assert len(ds1) == 1


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import numpy as np
import pytest

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from grib_fixtures import FL_NUMPY  # noqa: E402
from grib_fixtures import load_grib_data  # noqa: E402


@pytest.mark.parametrize("fl_type", FL_NUMPY)
def test_icon_to_xarray(fl_type):
    # test the conversion to xarray for an icon (unstructured grid) grib file.
    g, _ = load_grib_data("test_icon.grib", fl_type, folder="data")

    ds = g.to_xarray(engine="cfgrib")
    assert len(ds.data_vars) == 1
    # Dataset contains 9 levels and 9 grid points per level
    ref_levs = g.metadata("level")
    assert ds["pres"].sizes["generalVerticalLayer"] == len(ref_levs)
    assert ds["pres"].sizes["values"] == 6


@pytest.mark.parametrize("fl_type", FL_NUMPY)
def test_to_xarray_filter_by_keys(fl_type):
    g, _ = load_grib_data("tuv_pl.grib", fl_type)
    g = g.sel(param="t", level=500) + g.sel(param="u")
    assert len(g) > 1

    # see github #250
    r = g.to_xarray(
        engine="cfgrib", xarray_open_dataset_kwargs={"backend_kwargs": {"filter_by_keys": {"shortName": "t"}}}
    )

    assert len(r.data_vars) == 1
    assert r["t"].sizes["isobaricInhPa"] == 1


@pytest.mark.parametrize("fl_type", FL_NUMPY)
def test_grib_to_pandas(fl_type):
    f, _ = load_grib_data("test_single.grib", fl_type, folder="data")

    # all points
    df = f.to_pandas()
    assert len(df) == 84
    cols = [
        "lat",
        "lon",
        "value",
        "datetime",
        "domain",
        "levtype",
        "date",
        "time",
        "step",
        "param",
        "class",
        "type",
        "stream",
        "expver",
    ]
    assert list(df.columns) == cols
    assert np.allclose(df["lat"][0:2], [90, 90])
    assert np.allclose(df["lon"][0:2], [0, 30])
    assert np.allclose(df["value"][0:2], [260.435608, 260.435608])

    # specific location
    df = f.to_pandas(latitude=90, longitude=30)
    assert len(df) == 1
    assert list(df.columns) == cols
    assert np.isclose(df["lat"][0], 90)
    assert np.isclose(df["lon"][0], 30)
    assert np.isclose(df["value"][0], 260.435608)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import pytest
import yaml

from earthkit.data import FieldList
from earthkit.data import from_source
from earthkit.data.core.gridspec import GridSpec
from earthkit.data.core.temporary import temp_file
from earthkit.data.readers.grib.gridspec import GridSpecConverter
from earthkit.data.readers.grib.gridspec import make_gridspec
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import earthkit_remote_test_data_file
from earthkit.data.testing import earthkit_test_data_file
from earthkit.data.testing import write_to_file

SUPPORTED_GRID_TYPES = [
    "sh",
    "regular_ll",
    "reduced_ll",
    "regular_gg",
    "reduced_gg",
    "healpix",
]
UNSUPPORTED_GRID_TYPES = ["rotated_ll", "rotated_gg", "reduced_rotated_gg"]


def _make_gridspec_list(grid_types):
    d = []
    if isinstance(grid_types, str):
        grid_types = [grid_types]

    for gr in grid_types:
        with open(earthkit_test_data_file(os.path.join("gridspec", f"{gr}.yaml")), "r") as f:
            r = yaml.safe_load(f)
            for t in r:
                assert "file" in t, f"Missing 'file' key in {t}"
                assert "gridspec" in t, f"Missing 'gridspec' key in {t}"
                assert "metadata" in t, f"Missing 'metadata' key in {t}"
            d.extend(r)

    # print(d)
    return d


def gridspec_list(grid_types):
    if isinstance(grid_types, str):
        grid_types = [grid_types]

    for item in _make_gridspec_list(grid_types):
        yield (item["metadata"], item["gridspec"], item["file"])


# def gridspec_list_invalid(grid_types=None):
#     if grid_types is not None:
#         if isinstance(grid_types, str):
#             grid_types = [grid_types]
#     else:
#         grid_types = ["rotated_ll", "rotated_gg", "reduced_rotated_gg"]

#     for item in gridspec_list(grid_types):
#         yield (item["metadata"], item["gridspec"], item["file"])


@pytest.mark.parametrize(
    "metadata,ref,name",
    gridspec_list(SUPPORTED_GRID_TYPES),
)
def test_grib_gridspec_from_metadata_valid(metadata, ref, name):
    if name in [
        "regular_ll/wrf_swh_aegean_ll_jscanpos.grib1",
        "regular_ll/wind_uk_ll_jscanpos_jcons.grib1",
        # "regular_ll/t_global_0_360_5x5.grib1",
    ]:
        pytest.skip()

    gridspec = make_gridspec(metadata)
    assert dict(gridspec) == ref, name


@pytest.mark.parametrize(
    "metadata,ref,name",
    gridspec_list(UNSUPPORTED_GRID_TYPES),
)
def test_grib_gridspec_from_metadata_invalid_1(metadata, ref, name):
    with pytest.raises(ValueError):
        make_gridspec(metadata)


@pytest.mark.parametrize(
    "metadata",
    [
        {"gridType": "lambert"},
        {"gridType": "lambert_azimuthal_equal_area"},
        {"gridType": "mercator"},
        {"gridType": "polar_stereographic"},
    ],
)
def test_grib_gridspec_from_metadata_invalid_2(metadata):
    with pytest.raises(ValueError):
        make_gridspec(metadata)


def test_grib_gridspec_from_file():
    ds = from_source(
        "file",
        earthkit_test_data_file(os.path.join("gridspec", "t_75_-60_10_40_5x5.grib1")),
    )

    ref = {
        "type": "regular_ll",
        "grid": [5.0, 5.0],
        "area": [75.0, -60.0, 10.0, 40.0],
        "j_points_consecutive": 0,
        "i_scans_negatively": 0,
        "j_scans_positively": 0,
    }
    gs = ds[0].metadata().gridspec
    assert isinstance(gs, GridSpec)
    assert dict(gs) == ref


@pytest.mark.parametrize("metadata,gridspec,name", gridspec_list("regular_ll"))
def test_grib_metadata_from_gridspec_valid(metadata, gridspec, name):
    if name in [
        "regular_ll/wrf_swh_aegean_ll_jscanpos.grib1",
        "regular_ll/wind_uk_ll_jscanpos_jcons.grib1",
        # "regular_ll/t_global_0_360_5x5.grib1",
    ]:
        pytest.skip()

    edition = int(name[-1])
    assert edition in [1, 2]
    md, _ = GridSpecConverter.to_metadata(gridspec, edition=edition)
    assert md == metadata, name


@pytest.mark.parametrize(
    "metadata,gridspec,name",
    gridspec_list(
        [
            "sh",
            "reduced_ll",
            "regular_gg",
            "reduced_gg",
            "healpix",
        ]
    ),
)
def test_grib_metadata_from_gridspec_invalid(metadata, gridspec, name):
    if name in [
        "regular_ll/wrf_swh_aegean_ll_jscanpos.grib1",
        "regular_ll/wind_uk_ll_jscanpos_jcons.grib1",
        # "regular_ll/t_global_0_360_5x5.grib1",
    ]:
        pytest.skip()

    edition = int(name[-1])
    assert edition in [1, 2]
    with pytest.raises(ValueError):
        _, _ = GridSpecConverter.to_metadata(gridspec, edition=edition)


@pytest.mark.parametrize(
    "input_file",
    [
        "test-data/grids/reduced_gg/O32_global.grib1",
        "test-data/grids/reduced_gg/O32_global.grib2",
        "test-data/grids/healpix/H4_ring.grib2",
        "test-data/grids/healpix/H4_nested.grib2",
    ],
)
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_gridspec_to_fieldlist(input_file, write_method):
    import numpy as np

    def make_lat_lon(dx):
        lat_v = np.linspace(90, -90, int(180 / dx) + 1)
        lon_v = np.linspace(0, 360 - dx, int(360 / dx))
        lon, lat = np.meshgrid(lon_v, lat_v)
        return lat, lon

    ds_in = from_source(
        "url",
        earthkit_remote_test_data_file(input_file),
    )

    # target grid: 5x5
    gs = {"grid": [5, 5]}
    v = np.random.rand(37, 72)
    md = ds_in[0].metadata().override(gridspec=gs)
    ds = FieldList.from_numpy(v.flatten(), md)

    assert len(ds) == 1

    # make references
    ref = dict(
        Ni=72,
        Nj=37,
        Nx=72,
        Ny=37,
        gridType="regular_ll",
        iDirectionIncrementInDegrees=5.0,
        iScansNegatively=0,
        jDirectionIncrementInDegrees=5.0,
        jPointsAreConsecutive=0,
        jScansPositively=0,
        latitudeOfFirstGridPointInDegrees=90.0,
        latitudeOfLastGridPointInDegrees=-90.0,
        longitudeOfFirstGridPointInDegrees=0.0,
        longitudeOfLastGridPointInDegrees=355.0,
    )

    lat_ref, lon_ref = make_lat_lon(5)

    # check latlon
    latlon = ds[0].to_latlon()
    lat = latlon["lat"]
    lon = latlon["lon"]
    assert np.allclose(lat.flatten(), lat_ref.flatten())
    assert np.allclose(lon.flatten(), lon_ref.flatten())

    # check metadata
    res = {k: ds[0].metadata(k) for k in ref.keys()}
    assert res == ref

    # save
    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds)
        assert os.path.exists(tmp)
        r_tmp = from_source("file", tmp)

        # values
        v_tmp = r_tmp[0].values
        assert np.allclose(v_tmp, v.flatten(), atol=1e-3)

        # latlon
        latlon = r_tmp[0].to_latlon()
        lat = latlon["lat"]
        lon = latlon["lon"]
        assert np.allclose(lat.flatten(), lat_ref.flatten())
        assert np.allclose(lon.flatten(), lon_ref.flatten())

        v_tmp = r_tmp[0].values
        assert np.allclose(v_tmp, v.flatten(), atol=1e-3)
        res = {k: r_tmp[0].metadata(k) for k in ref.keys()}
        assert res == ref


def _global_grids():
    import json

    from earthkit.data.utils.paths import earthkit_conf_file

    with open(earthkit_conf_file("global_grids.json"), "r") as f:
        d = json.load(f)

    for _, v in d.items():
        yield v


# these are all the cases for the earthkit-regrid target grids
@pytest.mark.parametrize("edition", [1, 2])
@pytest.mark.parametrize("grid", _global_grids())
def test_grib_global_ll_gridspec_to_metadata(edition, grid):
    gs = {"grid": grid["grid"]}
    md, num = GridSpecConverter.to_metadata(gs, edition=edition)

    ref = dict(
        Ni=grid["shape"][1],
        Nj=grid["shape"][0],
        Nx=grid["shape"][1],
        Ny=grid["shape"][0],
        gridType="regular_ll",
        iDirectionIncrementInDegrees=grid["grid"][0],
        iScansNegatively=0,
        jDirectionIncrementInDegrees=grid["grid"][1],
        jPointsAreConsecutive=0,
        jScansPositively=0,
        latitudeOfFirstGridPointInDegrees=grid["area"][0],
        latitudeOfLastGridPointInDegrees=grid["area"][2],
        longitudeOfFirstGridPointInDegrees=grid["area"][1],
        longitudeOfLastGridPointInDegrees=grid["area"][3],
    )
    assert ref == md
    assert grid["shape"][0] * grid["shape"][1] == num


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from array_fl_fixtures import load_array_fl  # noqa: E402
from array_fl_fixtures import load_array_fl_file  # noqa: E402

# Note: Almost all grib metadata tests are also run for numpyfs.
# See grib/test_grib_metadata.py


def test_array_fl_field_repr():
    ds, _ = load_array_fl(1)

    t = repr(ds[0])
    assert t == "ArrayField(2t,None,20200513,1200,0,0)"


def test_array_fl_values_metadata_basic():
    ds, _ = load_array_fl(1)

    # values metadata
    # keys = [
    #     "min",
    #     "max",
    #     "avg",
    #     "ds",
    #     "skew",
    #     "kurt",
    #     "isConstant",
    #     "const",
    #     "bitmapPresent",
    #     "numberOfMissing",
    #     "values",
    # ]
    # for k in keys:
    #     assert ds[0].metadata(k, default=None) is None, k
    #     with pytest.raises(KeyError):
    #         ds[0].metadata(k)

    # bits per value must be kept from the original GRIB data
    assert ds[0].metadata("bitsPerValue") == 16


def test_array_fl_values_metadata_internal():
    ds, _ = load_array_fl(1)

    keys = {
        "shortName": "2t",
        "grib.shortName": "2t",
    }

    for k, v in keys.items():
        assert ds[0].metadata(k) == v, k


def test_array_fl_metadata_keys():
    ds, _ = load_array_fl(1)

    # The number/order of metadata keys can vary with the ecCodes version.
    # The same is true for the namespaces.

    md = ds[0].metadata()
    md_num = len(md)
    assert md_num > 100

    keys = md.keys()
    assert len(keys) == md_num

    for k in md.keys():
        assert isinstance(k, str)
        assert k != ""
        break

    items = md.items()
    assert len(list(items)) == md_num

    for k, v in md.items():
        assert isinstance(k, str)
        assert k != ""
        assert v is not None
        break

    assert "max" not in md
    assert "maximum" not in md
    assert "statistics.max" not in md
    assert "validityDate" in md


def test_array_fl_metadata_namespace():
    f, _ = load_array_fl_file("tuv_pl.grib")

    r = f[0].metadata(namespace="vertical")
    ref = {"level": 1000, "typeOfLevel": "isobaricInhPa"}
    assert r == ref

    r = f[0].metadata(namespace=["vertical", "time"])
    ref = {
        "vertical": {"typeOfLevel": "isobaricInhPa", "level": 1000},
        "time": {
            "dataDate": 20180801,
            "dataTime": 1200,
            "stepUnits": 1,
            "stepType": "instant",
            "stepRange": "0",
            "startStep": 0,
            "endStep": 0,
            "validityDate": 20180801,
            "validityTime": 1200,
        },
    }
    assert r == ref

    # The number/order of metadata keys can vary with the ecCodes version.
    # The same is true for the namespaces.

    r = f[0].metadata(namespace=None)
    assert isinstance(r, dict)
    md_num = len(r)
    assert md_num > 100
    assert r["level"] == 1000
    assert r["stepType"] == "instant"

    r = f[0].metadata(namespace=[None])
    assert isinstance(r, dict)
    assert len(r) == md_num
    assert r["level"] == 1000
    assert r["stepType"] == "instant"

    r = f[0].metadata(namespace="")
    assert isinstance(r, dict)
    assert len(r) == md_num
    assert r["level"] == 1000
    assert r["stepType"] == "instant"

    r = f[0].metadata(namespace=[""])
    assert isinstance(r, dict)
    assert len(r) == md_num
    assert r["level"] == 1000
    assert r["stepType"] == "instant"

    ref = {
        "geography",
        "vertical",
        "time",
        "parameter",
        "mars",
        "ls",
        "default",
    }
    r = f[0].metadata(namespace=all)
    assert isinstance(r, dict)
    assert set(r.keys()) == ref

    r = f[0].metadata(namespace=[all])
    assert isinstance(r, dict)
    assert set(r.keys()) == ref

    with pytest.raises(ValueError) as excinfo:
        r = f[0].metadata("level", namespace=["vertical", "time"])
    assert "must be a str when key specified" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        r = f[0].metadata("level", namespace=["vertical", "time"])
    assert "must be a str when key specified" in str(excinfo.value)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file
from earthkit.data.utils.xarray.profile import PROFILE_CONF

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from xr_engine_fixtures import compare_dims  # noqa: E402


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs,dims",
    [
        (
            {"profile": "mars", "level_dim_mode": "level", "dim_name_from_role_name": False},
            {"levelist": [300, 400, 500, 700, 850, 1000]},
        ),
        (
            {"profile": "mars", "level_dim_mode": "level_and_type", "dim_name_from_role_name": False},
            {"level_and_type": ["1000pl", "300pl", "400pl", "500pl", "700pl", "850pl"]},
        ),
    ],
)
def test_xr_level_dim(kwargs, dims):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))

    ds = ds_ek.to_xarray(**kwargs)
    compare_dims(ds, dims, order_ref_var="t")


@pytest.mark.cache
@pytest.mark.parametrize(
    "fname,kwargs,dims,levtype",
    [
        (
            "pl.grib",
            {"profile": "grib", "level_dim_mode": "level", "dim_name_from_role_name": False},
            {"level": [300, 400, 500, 700, 850, 1000]},
            "isobaricInhPa",
        ),
        (
            "pl_80_Pa.grib2",
            {
                "profile": "grib",
                "level_dim_mode": "level",
                "ensure_dims": "level",
                "dim_name_from_role_name": False,
            },
            {"level": [80]},
            "isobaricInPa",
        ),
        (
            "hpa_and_pa.grib",
            {
                "profile": "mars",
                "level_dim_mode": "level",
                "ensure_dims": "levelist",
                "dim_name_from_role_name": False,
            },
            {"levelist": [0.01, 0.1, 1]},
            "pl",
        ),
        (
            "hl_1000_m_asl.grib2",
            {
                "profile": "grib",
                "level_dim_mode": "level",
                "ensure_dims": "level",
                "dim_name_from_role_name": False,
            },
            {"level": [100, 1000, 2000, 3000]},
            "heightAboveSea",
        ),
        (
            "hl_1000_m_agr.grib2",
            {
                "profile": "grib",
                "level_dim_mode": "level",
                "ensure_dims": "level",
                "dim_name_from_role_name": False,
            },
            {"level": [500, 1000, 2500, 10000]},
            "heightAboveGround",
        ),
        (
            "pt_320_K.grib1",
            {
                "profile": "grib",
                "level_dim_mode": "level",
                "ensure_dims": "level",
                "dim_name_from_role_name": False,
            },
            {"level": [320]},
            "theta",
        ),
        (
            "pv_1500.grib1",
            {
                "profile": "grib",
                "level_dim_mode": "level",
                "ensure_dims": "level",
                "dim_name_from_role_name": False,
            },
            {"level": [1500]},
            "potentialVorticity",
        ),
        (
            "soil_7.grib1",
            {
                "profile": "grib",
                "level_dim_mode": "level",
                "ensure_dims": "level",
                "dim_name_from_role_name": False,
            },
            {"level": [7]},
            "depthBelowLand",
        ),
        (
            "sol_3.grib2",
            {
                "profile": "grib",
                "level_dim_mode": "level",
                "ensure_dims": "level",
                "dim_name_from_role_name": False,
            },
            {"level": [3]},
            "snowLayer",
        ),
        (
            "ml_77.grib2",
            {
                "profile": "grib",
                "level_dim_mode": "level",
                "ensure_dims": "level",
                "dim_name_from_role_name": False,
            },
            {"level": [77]},
            "hybrid",
        ),
        (
            "sfc.grib1",
            {
                "profile": "grib",
                "level_dim_mode": "level",
                "ensure_dims": "level",
                "dim_name_from_role_name": False,
            },
            {"level": [0]},
            "surface",
        ),
        # (
        #     "sfc.grib2",
        #     {"profile": "grib", "level_dim_mode": "level", "ensure_dims": "level"},
        #     {"level": [0]},
        #     "surface",
        # ),
        (
            "mean_sea_level_reduced_ll.grib1",
            {
                "profile": "grib",
                "level_dim_mode": "level",
                "ensure_dims": "level",
                "dim_name_from_role_name": False,
            },
            {"level": [0]},
            "meanSea",
        ),
        (
            "gen_vert_layer.grib",
            {
                "profile": "grib",
                "level_dim_mode": "level",
                "ensure_dims": "level",
                "dim_name_from_role_name": False,
            },
            {"level": [1]},
            "generalVerticalLayer",
        ),
        (
            "pl.grib",
            {"profile": "mars", "level_dim_mode": "level", "dim_name_from_role_name": False},
            {"levelist": [300, 400, 500, 700, 850, 1000]},
            "pl",
        ),
        (
            "pl_80_Pa.grib2",
            {
                "profile": "mars",
                "level_dim_mode": "level",
                "ensure_dims": "levelist",
                "dim_name_from_role_name": False,
            },
            {"levelist": [0.8]},
            "pl",
        ),
        (
            "pt_320_K.grib1",
            {
                "profile": "mars",
                "level_dim_mode": "level",
                "ensure_dims": "levelist",
                "dim_name_from_role_name": False,
            },
            {"levelist": [320]},
            "pt",
        ),
        (
            "pv_1500.grib1",
            {
                "profile": "mars",
                "level_dim_mode": "level",
                "ensure_dims": "levelist",
                "dim_name_from_role_name": False,
            },
            {"levelist": [1500]},
            "pv",
        ),
        (
            "sol_3.grib2",
            {
                "profile": "grib",
                "level_dim_mode": "level",
                "ensure_dims": "levelist",
                "dim_name_from_role_name": False,
            },
            {"levelist": [3]},
            "sol",
        ),
        (
            "hpa_and_pa.grib",
            {
                "profile": "mars",
                "level_dim_mode": "level",
                "ensure_dims": "levelist",
                "dim_name_from_role_name": False,
            },
            {"levelist": [0.01, 0.1, 1]},
            "pl",
        ),
    ],
)
def test_xr_level_attr(fname, kwargs, dims, levtype):
    ds_ek = from_source("url", earthkit_remote_test_data_file(f"test-data/xr_engine/level/{fname}"))

    ds = ds_ek.to_xarray(**kwargs)
    compare_dims(ds, dims)

    level_dim = next(iter(dims))
    ds.coords[level_dim].attrs == PROFILE_CONF.defaults["coord_attrs"][level_dim][levtype]



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import sys

import numpy as np
import pytest
from earthkit.utils.testing import check_array_type

from earthkit.data import from_source
from earthkit.data.core.fieldlist import FieldList
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import ARRAY_BACKENDS
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_test_data_file
from earthkit.data.testing import write_to_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from array_fl_fixtures import load_array_fl  # noqa: E402

LOG = logging.getLogger(__name__)


@pytest.mark.parametrize("array_backend", ARRAY_BACKENDS)
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_array_fl_grib_write_to_path(array_backend, write_method):

    ds = from_source("file", earthkit_examples_file("test.grib"))
    ds = ds.to_fieldlist(array_backend=array_backend.name)

    assert ds[0].metadata("shortName") == "2t"
    assert len(ds) == 2
    v1 = ds[0].values + 1
    check_array_type(v1, array_backend)

    md = ds[0].metadata()
    md1 = md.override(shortName="msl")
    r = FieldList.from_array(v1, md1)

    with temp_file() as tmp:
        write_to_file(write_method, tmp, r)
        assert os.path.exists(tmp)
        r_tmp = from_source("file", tmp)
        r_tmp = r_tmp.to_fieldlist(array_backend=array_backend.name)
        v_tmp = r_tmp[0].values
        assert array_backend.allclose(v1, v_tmp)


@pytest.mark.parametrize("array_backend", ARRAY_BACKENDS)
@pytest.mark.parametrize("_kwargs", [{}, {"check_nans": True}])
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_array_fl_grib_write_missing(array_backend, _kwargs, write_method):

    ds = from_source("file", earthkit_examples_file("test.grib"))
    ds = ds.to_fieldlist(array_backend=array_backend.name)
    # ns = get_array_namespace(array_backend)

    ns = array_backend.compat_namespace

    assert ds[0].metadata("shortName") == "2t"

    v = ds[0].values
    v1 = v + 1
    assert not ns.isnan(v1[0])
    assert not ns.isnan(v1[1])
    v1[0] = ns.nan
    assert ns.isnan(v1[0])
    assert not ns.isnan(v1[1])

    md = ds[0].metadata()
    md1 = md.override(shortName="msl")
    r = FieldList.from_array(v1, md1)

    assert ns.isnan(r[0].values[0])
    assert not ns.isnan(r[0].values[1])

    with temp_file() as tmp:
        write_to_file(write_method, tmp, r, **_kwargs)
        assert os.path.exists(tmp)
        r_tmp = from_source("file", tmp)
        v_tmp = r_tmp[0].values
        assert np.isnan(v_tmp[0])
        r_tmp = r_tmp.to_fieldlist(array_backend=array_backend.name)
        v_tmp = r_tmp[0].values
        assert ns.isnan(v_tmp[0])
        assert not ns.isnan(v_tmp[1])


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_array_fl_grib_write_check_nans_bad(write_method):
    ds = from_source("file", earthkit_examples_file("test.grib"))

    assert ds[0].metadata("shortName") == "2t"

    v = ds[0].values
    v1 = v + 1
    assert not np.isnan(v1[0])
    assert not np.isnan(v1[1])
    v1[0] = np.nan
    assert np.isnan(v1[0])
    assert not np.isnan(v1[1])

    md = ds[0].metadata()
    md1 = md.override(shortName="msl")
    r = FieldList.from_numpy(v1, md1)

    assert np.isnan(r[0].values[0])
    assert not np.isnan(r[0].values[1])

    with temp_file() as tmp:
        from eccodes import EncodingError

        with pytest.raises(EncodingError):
            write_to_file(write_method, tmp, r, check_nans=False)


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_array_fl_grib_write_append(write_method):
    ds = from_source("file", earthkit_examples_file("test.grib"))

    assert ds[0].metadata("shortName") == "2t"

    v = ds[0].values
    v1 = v + 1
    v2 = v + 2

    md = ds[0].metadata()
    md1 = md.override(shortName="msl")
    md2 = md.override(shortName="2d")

    r1 = FieldList.from_numpy(v1, md1)
    r2 = FieldList.from_numpy(v2, md2)

    # save to disk
    tmp = temp_file()
    write_to_file(write_method, tmp.path, r1)
    assert os.path.exists(tmp.path)
    r_tmp = from_source("file", tmp.path)
    assert len(r_tmp) == 1
    assert r_tmp.metadata("shortName") == ["msl"]
    r_tmp = None

    # append
    write_to_file(write_method, tmp.path, r2, append=True)
    assert os.path.exists(tmp.path)
    r_tmp = from_source("file", tmp.path)
    assert len(r_tmp) == 2
    assert r_tmp.metadata("shortName") == ["msl", "2d"]


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_array_fl_grib_write_generating_proc_id(write_method):
    ds = from_source("file", earthkit_examples_file("test.grib"))

    assert ds[0].metadata("shortName") == "2t"

    v = ds[0].values
    v1 = v + 1
    v2 = v + 2

    md = ds[0].metadata()
    md1 = md.override(shortName="msl", generatingProcessIdentifier=255)
    md2 = md.override(shortName="2d")

    r1 = FieldList.from_numpy([v1, v2], [md1, md2])

    # save to disk
    with temp_file() as tmp:
        write_to_file(write_method, tmp, r1)
        assert os.path.exists(tmp)
        r_tmp = from_source("file", tmp)
        assert len(r_tmp) == 2
        assert r_tmp.metadata("shortName") == ["msl", "2d"]
        assert r_tmp.metadata("generatingProcessIdentifier") == [
            255,
            150,
        ]

        assert np.allclose(r_tmp.values[0], v1)
        assert np.allclose(r_tmp.values[1], v2)


@pytest.mark.parametrize("array_backend", ARRAY_BACKENDS)
@pytest.mark.parametrize(
    "_kwargs,expected_value",
    [({}, None), ({"bits_per_value": 12}, 12), ({"bits_per_value": None}, None)],
)
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_array_fl_grib_write_bits_per_value(array_backend, _kwargs, expected_value, write_method):
    ds, _ = load_array_fl(1, array_backend)

    if expected_value is None:
        expected_value = ds[0].metadata("bitsPerValue")

    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds, **_kwargs)
        ds1 = from_source("file", tmp)
        assert ds1.metadata("bitsPerValue") == [expected_value] * len(ds)


@pytest.mark.parametrize(
    "filename,shape",
    [
        (earthkit_examples_file("test.grib"), (11, 19)),
        (earthkit_test_data_file("O32_global.grib1"), (5248,)),
        (earthkit_test_data_file("O32_global.grib2"), (5248,)),
    ],
)
def test_array_fl_grib_single_write_to_path(filename, shape):
    ds = from_source("file", filename)

    assert len(ds) >= 1
    v1 = ds[0].values + 1

    md = ds[0].metadata()
    md1 = md.override(shortName="msl")
    r = FieldList.from_array(v1, md1)
    assert r[0].shape == shape

    with temp_file() as tmp:
        r.save(tmp)
        assert os.path.exists(tmp)
        r_tmp = from_source("file", tmp)
        # r_tmp = r_tmp.to_fieldlist(array_backend=array_backend)
        assert r_tmp[0].shape == shape
        assert r_tmp[0].metadata("shortName") == "msl"
        v_tmp = r_tmp[0].values
        assert np.allclose(v1, v_tmp)


@pytest.mark.parametrize(
    "filename,shape",
    [
        (earthkit_examples_file("test.grib"), (11, 19)),
        (earthkit_test_data_file("O32_global.grib1"), (5248,)),
        (earthkit_test_data_file("O32_global.grib2"), (5248,)),
    ],
)
@pytest.mark.parametrize(
    "_kwargs,expected_value",
    [({}, None), ({"bits_per_value": 8}, 8), ({"bits_per_value": None}, None)],
)
def test_array_fl_grib_single_write_bits_per_value(filename, shape, _kwargs, expected_value):
    ds0 = from_source("file", filename)

    ds = ds0.from_fields([ds0[0].copy()])
    assert ds[0].shape == shape

    if expected_value is None:
        expected_value = ds[0].metadata("bitsPerValue")

    with temp_file() as tmp:
        ds.save(tmp, **_kwargs)
        ds1 = from_source("file", tmp)
        assert ds1.metadata("bitsPerValue") == [expected_value] * len(ds)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file


# TODO: the disabled tests require a fix in multiurl
@pytest.mark.parametrize(
    "parts,expected_meta",
    [
        ([(0, 150)], [("t", 1000)]),  # message length
        ([(0, 154)], [("t", 1000)]),  # message length + extra bytes
        ([(0, 240)], [("t", 1000)]),  # message length + padding
        (
            [(0, 240)],
            [("t", 1000)],
        ),  # message length + padding + bytes from next message
        # ([(0, 15)], []),  # shorter than message
        ([(240, 150)], [("u", 1000)]),
        ([(240, 480)], [("u", 1000), ("v", 1000)]),
        ([(240, 240), (720, 240)], [("u", 1000), ("t", 850)]),
        ([(240, 240), (720, 243)], [("u", 1000), ("t", 850)]),
        # ([(240, 240), (720, 16)], [("u", 1000)]),  # second part shorter than message
    ],
)
def test_grib_single_url_parts(parts, expected_meta):
    ds = from_source("url", earthkit_remote_test_data_file("examples/test6.grib"), parts=parts)

    assert len(ds) == len(expected_meta)
    if len(ds) > 0:
        assert ds.metadata(("param", "level")) == expected_meta


@pytest.mark.parametrize(
    "parts,expected_meta",
    [
        ([(0, 150)], [("t", 1000)]),
        ([(240, 150)], [("u", 1000)]),
        ([(240, 480)], [("u", 1000), ("v", 1000)]),
    ],
)
def test_grib_single_url_parts_1(parts, expected_meta):
    ds = from_source("url", [earthkit_remote_test_data_file("examples/test6.grib"), parts])

    assert len(ds) == len(expected_meta)

    cnt = 0
    for i, f in enumerate(ds):
        assert f.metadata(("param", "level")) == expected_meta[i], i
        cnt += 1

    assert cnt == len(expected_meta)


@pytest.mark.parametrize(
    "parts1,parts2,expected_meta",
    [
        (
            [(240, 150)],
            None,
            [("u", 1000), ("2t", 0), ("msl", 0)],
        ),
        (
            None,
            [(0, 526)],
            [
                ("t", 1000),
                ("u", 1000),
                ("v", 1000),
                ("t", 850),
                ("u", 850),
                ("v", 850),
                ("2t", 0),
            ],
        ),
        (
            [(240, 150)],
            [(0, 526)],
            [("u", 1000), ("2t", 0)],
        ),
    ],
)
def test_grib_multi_url_parts(parts1, parts2, expected_meta):
    ds = from_source(
        "url",
        [
            [earthkit_remote_test_data_file("examples/test6.grib"), parts1],
            [earthkit_remote_test_data_file("examples/test.grib"), parts2],
        ],
    )

    assert len(ds) == len(expected_meta)
    assert ds.metadata(("param", "level")) == expected_meta


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.testing import NO_HDA

CDS_TIMEOUT = pytest.CDS_TIMEOUT


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_HDA, reason="No access to WEKEO")
@pytest.mark.timeout(CDS_TIMEOUT)
@pytest.mark.parametrize("prompt", [True, False])
def test_wekeo_grib_1_prompt(prompt):
    s = from_source(
        "wekeocds",
        "EO:ECMWF:DAT:REANALYSIS_ERA5_SINGLE_LEVELS_MONTHLY_MEANS",
        variable=["2m_temperature", "mean_sea_level_pressure"],
        product_type=["monthly_averaged_reanalysis_by_hour_of_day"],
        year=["2012"],
        month=["12"],
        time=["13:00"],
        prompt=prompt,
        data_format="grib",
        download_format="zip",
    )
    assert len(s) == 2


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_HDA, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_wekeo_grib_2():
    s = from_source(
        "wekeocds",
        "EO:ECMWF:DAT:REANALYSIS_ERA5_SINGLE_LEVELS_MONTHLY_MEANS",
        variable=["2m_temperature", "mean_sea_level_pressure"],
        product_type=["monthly_averaged_reanalysis_by_hour_of_day"],
        year=["2012"],
        month=["12"],
        time=["13:00"],
        data_format="grib",
        download_format="zip",
        split_on="variable",
    )
    assert len(s) == 2


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_HDA, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_wekeo_grib_3():
    s = from_source(
        "wekeocds",
        "EO:ECMWF:DAT:REANALYSIS_ERA5_SINGLE_LEVELS_MONTHLY_MEANS",
        variable=["2m_temperature", "mean_sea_level_pressure"],
        product_type=["monthly_averaged_reanalysis_by_hour_of_day"],
        year=["2012"],
        month=["12"],
        time=["13:00"],
        data_format="grib",
        download_format="zip",
    )
    assert len(s) == 2


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_HDA, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_wekeo_netcdf():
    s = from_source(
        "wekeocds",
        "EO:ECMWF:DAT:REANALYSIS_ERA5_SINGLE_LEVELS_MONTHLY_MEANS",
        variable=["2m_temperature", "mean_sea_level_pressure"],
        product_type=["monthly_averaged_reanalysis_by_hour_of_day"],
        year=["2012"],
        month=["12"],
        time=["13:00"],
        data_format="netcdf",
        download_format="zip",
    )
    assert len(s) == 2


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)
